-- Coding standards brought to you by the East India Trading Company in association with the Save the Camel Foundation
local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local SM = LibStub("LibSharedMedia-3.0")
local TalentQuery = LibStub("LibTalentQuery-1.0")
local commPrefix = "Utopia"
local playerName, playerClass, db, title
local new, del, copy, deepDel, Gradient, UnitFullName, ClassColour, SmoothColour, SmoothColourStr, rotate
local petClasses = {HUNTER = true, WARLOCK = true}
local throttle
local specChangers = {
	[GetSpellInfo(63645) or "Fake"] = true,			-- Activate Primary Spec
	[GetSpellInfo(63644) or "Fake"] = true,			-- Activate Secondary Spec
}

local function d(...)
	if (ZDEBUG) then
		ChatFrame1:AddMessage("|cFFFF8080DEBUG:|r "..format(...), 0.8, 0.8, 0.8)
	end
end

assert(AceGUIWidgetLSMlists, "Utopia requires AceGUI-3.0-SharedMediaWidgets")

local module = LibStub("AceAddon-3.0"):NewAddon("Utopia", "AceConsole-3.0", "AceHook-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0")
--@debug@
Utopia = module
--@no-debug@
module.version = tonumber(strmatch("$Revision: 83 $", "(%d+)")) or 0
function Utopia:UpdateVersion(ver)
	self.version = max(tonumber(strmatch(ver, "(%d+)")) or 0, self.version)
end

local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("Utopia", {
	type = "data source",
	label = L["TITLE"],
	icon = select(3, GetSpellInfo(47467)),
})

-- LDB Bits
function LDB:Update()
	self.label = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
end

-- OnClick
function LDB.OnClick(self, button)
	if (not IsModifierKeyDown()) then
		if (button == "LeftButton") then
			module:ToggleActive()
		elseif (button == "RightButton") then
			module:OpenConfig()
		end
	end
end

-- OnTooltipShow
function LDB.OnTooltipShow(tt)
	tt:AddLine(module.title)

	local gotTalents = module:GetTalentCount()
	local missing = module:GetTalentMissingPeople()

	local raidSize = GetNumRaidMembers()
	if (raidSize == 0) then
		raidSize = GetNumPartyMembers() + 1
	end

	local r, g, b = SmoothColour(gotTalents / raidSize)
	tt:AddDoubleLine(L["Group size:"], tostring(raidSize), 1, 1, 1, 0, 1, 0)
	tt:AddDoubleLine(L["Got Talents for:"], tostring(gotTalents), 1, 1, 1, r, g, b)
	if (missing) then
		tt:AddLine(format(L["Missing: %s"], missing), 1, 1, 1, 1)
	end
end

module.options = {
	type = 'group',
	order = 15,
	name = L["TITLE"],
	desc = L["Configuration"],
	handler = module,
	args = {
		General = {
			type = 'group',
			order = 1,
			name = L["General Settings"],
			desc = L["General Settings"],
			args = {
				enable = {
					type = 'toggle',
					name = L["Enable"],
					desc = L["Enable target debuff status display"],
					get = function() return db.enable end,
					set = function(info, n) db.enable = n module:ShowWhen() end,
					order = 5,
				},
				locked = {
					type = 'toggle',
					name = L["Locked"],
					desc = L["Toggle locked position. When unlocked, you can move the display with the bar"],
					get = function() return db.locked end,
					set = function(info, n)
						db.locked = n
						if (module.frame.bar) then
							module.frame.bar:EnableMouse(not db.locked)
						end

						for i,icon in pairs(module.frame.icons) do
							if (n) then
								icon:RegisterForDrag(nil)
							else
								icon:RegisterForDrag("LeftButton")		-- ,RightButton
							end
						end

						if (module.frame.backdrop) then
							module.frame.backdrop:EnableMouse(not db.locked)
						end
					end,
					order = 20,
				},
--@debug@
				pvp = {
					type = 'toggle',
					name = "PvP (debug only)",
					desc = "Enable in pvp instances",
					get = function() return module.pvp end,
					set = function(info, n) module.pvp = n module:ShowWhen() end,
					order = 25,
				},
--@no-debug@
				mode = {
					type = 'group',
					name = L["Mode Options"],
					desc = L["Mode Options"],
					guiInline = true,
					get = function(info) return db[info[#info]] end,
					set = function(info, value) db[info[#info]] = value end,
					order = 100,
					args = {
						automode = {
							type = 'toggle',
							name = L["Auto Mode"],
							desc = L["Automatically choose buffs mode out of combat and Debuffs mode in combat"],
							order = 5,
						},
						allPlayers = {
							type = 'toggle',
							name = L["All Players"],
							desc = L["Targetting players will show their buff status instead of your own"],
							order = 10,
						},
					},
				},
				tooltip = {
					type = 'group',
					name = L["Tooltip Options"],
					desc = L["Options for default values when unavaible"],
					guiInline = true,
					get = function(info) return db[info[#info]] end,
					set = function(info, value) db[info[#info]] = value end,
					order = 150,
					args = {
						defaultMax = {
							type = 'toggle',
							name = L["Max Possible"],
							desc = L["Unavailable spells are shown as max possible talented value, instead of the untalented base value"],
							order = 5,
						},
						totemsTemporary = {
							type = 'toggle',
							name = L["Totems are Temporary"],
							desc = L["Treat all totem buffs as if they are temporary and not important pre-combat."],
							order = 10,
							set = function(info, value)
								db[info[#info]] = value
								if (module.frame) then
									module.frame.lastTargetGUID = nil
									module.frame:SetMode()
								end
							end,
						},
					},
				},
				talent = {
					type = 'group',
					name = L["Talent Options"],
					desc = L["Talent Options"],
					guiInline = true,
					get = function(info) return db[info[#info]] end,
					set = function(info, value) db[info[#info]] = value end,
					order = 170,
					args = {
						storeGuildTalents = {
							type = 'toggle',
							name = L["Store Guild Members"],
							desc = L["Store last seen guild member talents for faster initial display. Talents are still refreshed when the members are seen. Talents are only stored when you are in a raid, as it's more likely for players to use alternate specs when in a party."],
							order = 10,
						},
						clear = {
							type = 'execute',
							name = L["Clear Guild Members"],
							desc = L["Clear stored guild member talents."],
							order = 15,
							disabled = function() return not db.guildTalents end,
							func = function()
								db.guildTalents = deepDel(db.guildTalents)
							end,
						},
						reset = {
							type = 'execute',
							name = L["Rescan Talents"],
							desc = L["Clear talent cache and rescan everyone"],
							func = function()
								module.talents = deepDel(module.talents)
								module.talents = new()
								Utopia:ScanRoster()
							end,
						},
					},
				},
				misc = {
					type = 'group',
					name = L["Miscelanious Options"],
					desc = L["Other options"],
					guiInline = true,
					get = function(info) return db[info[#info]] end,
					set = function(info, value) db[info[#info]] = value end,
					order = 200,
					args = {
						tooltip = {
							type = 'toggle',
							name = L["Tooltip Enhancements"],
							desc = L["Improve the information shown in default buff tooltips to reflect the talent enchanced amounts and show extra bonus affects granted."],
							order = 5,
							set = function(info, value)
								db[info[#info]] = value 
								module:CheckXPerlTooltipNamesOption()
							end,
						},
						tooltiprank = {
							type = 'toggle',
							name = L["Spell Rank"],
							desc = L["Show spell rank in standard tooltips."],
							order = 10,
						},
					},
				},
			},
		},
		icons = {
			type = 'group',
			name = L["Icons"],
			desc = L["Icon Settings"],
			order = 5,
			args = {
				auto = {
					type = "toggle",
					order = 1,
					name = L["Automatic"],
					desc = L["Automatically show the buffs relevant for your class"],
					get = function() return db.automatic end,
					set = function(info, value)
						db.automatic = value
						if (module.frame) then
							module.frame.lastTargetGUID = nil
							module.frame:SetMode()
						end
					end,
				},
				onlyAvailable = {
					type = "toggle",
					order = 1,
					name = L["Only Available"],
					desc = L["Only show available icons in the set, hiding all icons your group cannot apply buffs or debuffs for"],
					get = function() return db.onlyAvailable end,
					set = function(info, value)
						db.onlyAvailable = value
						if (module.frame) then
							module.frame.lastTargetGUID = nil
							module.frame:SetMode()
						end
					end,
				},
				buffs = {
					type = 'group',
					order = 100,
					name = L["Buffs"],
					desc = L["Buff Icon Options"],
					disabled = function() return db.automatic end,
					guiInline = true,
					get = function(info)
						local self = module.options.args.icons.args.buffs.args[info[#info]]
						return not (db.disabled and db.disabled.buffs and db.disabled.buffs[self.name])
					end,
					set = function(info, value)
						local self = module.options.args.icons.args.buffs.args[info[#info]]
						value = not value
						if (value) then
							if (not db.disabled) then
								db.disabled = new()
							end
							if (not db.disabled.buffs) then
								db.disabled.buffs = new()
							end
							db.disabled.buffs[self.name] = true
						else
							db.disabled.buffs[self.name] = nil
							if (not next(db.disabled.buffs)) then
								db.disabled.buffs = del(db.disabled.buffs)
								if (not next(db.disabled)) then
									db.disabled = del(db.disabled)
								end
							end
						end
						if (module.frame) then
							module.frame:SetMode("buffs")
						end
					end,
					args = {},		-- Filled later
				},
				debuffs = {
					type = 'group',
					order = 200,
					name = L["Debuffs"],
					desc = L["Debuff Icon Options"],
					disabled = function() return db.automatic end,
					guiInline = true,
					get = function(info)
						local self = module.options.args.icons.args.debuffs.args[info[#info]]
						return not (db.disabled and db.disabled.debuffs and db.disabled.debuffs[self.name])
					end,
					set = function(info, value)
						local self = module.options.args.icons.args.debuffs.args[info[#info]]

						value = not value
						if (value) then
							if (not db.disabled) then
								db.disabled = new()
							end
							if (not db.disabled.debuffs) then
								db.disabled.debuffs = new()
							end
							db.disabled.debuffs[self.name] = true
						else
							db.disabled.debuffs[self.name] = nil
							if (not next(db.disabled.debuffs)) then
								db.disabled.debuffs = del(db.disabled.debuffs)
								if (not next(db.disabled)) then
									db.disabled = del(db.disabled)
								end
							end
						end
						if (module.frame) then
							module.frame:SetMode("debuffs")
						end
					end,
					args = {},		-- Filled later
				}
			}
		},
		display = {
			type = 'group',
			name = L["Display"],
			desc = L["Display settings"],
			order = 10,
			args = {
				bar = {
					type = 'group',
					order = 100,
					name = L["Bar"],
					desc = L["Display options for the status bar"],
					guiInline = true,
					args = {
						Enable = {
							type = "toggle",
							order = 1,
							name = L["Enable"],
							desc = L["Enable the status bar"],
							get = function() return db.bar end,
							set = "EnableBar",
						},
						side = {
							type = 'select',
							name = L["Bar Placement"],
							desc = L["Which side to put the bar"],
							values = {B = L["Bottom/Right"], T = L["Top/Left"]},
							order = 10,
							get = function() return db.side end,
							set = "SetOrientation",
						},
						size = {
							type = "range",
							name = L["Icon Size"],
							desc = L["Set the size of the icons"],
							func = timeFunc,
							order = 2,
							get = function() return db.size end,
							set = "SetIconSize",
							min = 8,
							max = 64,
							step = 1,
							order = 20,
						},
						scale = {
							type = "range",
							name = L["Scale"],
							desc = L["Set the scale of the icons"],
							func = timeFunc,
							order = 2,
							get = function() return db.scale end,
							set = function(info, n) db.scale = n module:SetScale() module:SavePosition(module.frame) end,
							min = 0.2,
							max = 2,
							step = 0.01,
							bigStep = 0.1,
							order = 30,
						},
						texture = {
							type = 'select', dialogControl = 'LSM30_Statusbar',
							name = L["Bar Texture"],
							desc = L["Set the texture for the bar"],
							values = AceGUIWidgetLSMlists.statusbar,
							order = 40,
							get = function() return db.texture end,
							set = function(v,n)
								db.texture = n
								local tex = SM:Fetch("statusbar", n)
								if (module.frame.bar) then
									module.frame.bar:SetStatusBarTexture(tex)
									module.frame.bar.bg:SetTexture(tex)
								end
							end,
						},
						barwidth = {
							type = "range",
							name = L["Width"],
							desc = L["Set the width of the status bar"],
							func = timeFunc,
							order = 2,
							get = function() return db.barwidth end,
							set = function(info, n) db.barwidth = n module:SetOrientation() end,
							min = 1,
							max = 20,
							step = 1,
							order = 50,
						},
					},
				},
				buttons = {
					type = 'group',
					name = L["Buttons"],
					order = 200,
					guiInline = true,
					get = function(info) return db[info[#info]] end,
					set = function(info, value)
						db[info[#info]] = value
						module:SetOrientation()
					end,
					args = {
						orientation = {
							type = 'select',
							name = L["Orientation"],
							desc = L["Set the orientation for the display"],
							values = {V = L["Vertical"], H = L["Horizontal"]},
							order = 1,
							get = function() return db.orientation end,
							set = "SetOrientation",
						},
						hspacing = {
							type = "range",
							name = L["Horizontal Spacing"],
							desc = L["Set the horizontal spacing of the buttons"],
							func = timeFunc,
							order = 2,
							min = 0,
							max = 20,
							step = 1,
							order = 5,
						},
						vspacing = {
							type = "range",
							name = L["Vertical Spacing"],
							desc = L["Set the vertical spacing of the buttons"],
							func = timeFunc,
							order = 2,
							min = 0,
							max = 20,
							step = 1,
							order = 10,
						},
						columns = {
							type = "range",
							name = L["Columns"],
							desc = L["Set the buttons before breaking to next row/column"],
							func = timeFunc,
							order = 2,
							min = 1,
							max = 15,
							step = 1,
							order = 20,
						},
						popupPos = {
							type = 'select',
							name = L["Popup Toggle"],
							desc = L["Select the position for the popup toggle button"],
							values = {top = L["Top"], bottom = L["Bottom"]},
							order = 50,
						},
					},
				},
				colours = {
					type = 'group',
					name = L["Colour"],
					order = 300,
					guiInline = true,
					get = function(info)
						local c = db.colour[info[#info]]
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)		--value)
						local c = db.colour[info[#info]]
						c.r, c.g, c.b, c.a = r, g, b, a
						if (module.frame) then
							module.frame:UpdateBuffs()
						end
					end,
					args = {
						active = {
							type = "color",
							name = L["Active Colour"],
							desc = L["Set the colour for debuffs which your group has applied."],
							hasAlpha = true,
							order = 1,
						},
						inactive = {
							type = "color",
							name = L["Inactive Colour"],
							desc = L["Set the colour for debuffs which your group is able to apply, but has not done so."],
							hasAlpha = true,
							order = 2,
						},
          				partActive = {
							type = "color",
							name = L["Partial Colour"],
							desc = L["Set the colour for auras which someone has applied, but it is an unimproved version of the spell."],
							hasAlpha = true,
							order = 4,
						},
          				unavailable = {
							type = "color",
							name = L["Unavailable Colour"],
							desc = L["Set the colour for debuffs which your group is not able to apply."],
							hasAlpha = true,
							order = 4,
						},
          				excluded = {
							type = "color",
							name = L["Excluded Colour"],
							desc = L["Set the colour for debuffs which your group is able to apply, but another spell is excluding it."],
							hasAlpha = true,
							order = 5,
						},
						temporary = {
							type = "color",
							name = L["Temporary Colour"],
							desc = L["Set the colour for debuffs which your group is able to apply, but usually only during combat. They're expected to not be active when idle."],
							hasAlpha = true,
							order = 6,
						}
					},
				},
				background = {
					type = "group",
					name = L["Background Options"],
					guiInline = true,
					order = 400,
					get = function(info) return db.Background[ info[#info] ] end,
					set = function(info, value)
						db.Background[ info[#info] ] = value
						module:UpdateBackdrop()
					end,
					args = {
						Enable = {
							type = "toggle",
							order = 1,
							name = L["Enable"],
							desc = L["Enable the frame background"],
						},
						Texture = {
							type = "select", dialogControl = 'LSM30_Background',
							order = 10,
							name = L["Background Texture"],
							desc = L["Texture to use for the frame's background"],
							values = AceGUIWidgetLSMlists.background,
						},
						BorderTexture = {
							type = "select", dialogControl = 'LSM30_Border',
							order = 20,
							name = L["Border Texture"],
							desc = L["Texture to use for the frame's border"],
							values = AceGUIWidgetLSMlists.border,
						},
						Color = {
							type = "color",
							order = 30,
							name = L["Background Color"],
							desc = L["Frame's background color"],
							hasAlpha = true,
							get = function(info)
								local t = db.Background.Color
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = db.Background.Color
								t.r, t.g, t.b, t.a = r, g, b, a
								module:UpdateBackdrop()
							end,
						},
						BorderColor = {
							type = "color",
							order = 40,
							name = L["Border Color"],
							desc = L["Frame's border color"],
							hasAlpha = true,
							get = function(info)
								local t = db.Background.BorderColor
								return t.r, t.g, t.b, t.a
							end,
							set = function(info, r, g, b, a)
								local t = db.Background.BorderColor
								t.r, t.g, t.b, t.a = r, g, b, a
								module:UpdateBackdrop()
							end,
						},
						Tile = {
							type = "toggle",
							order = 50,
							name = L["Tile Background"],
							desc = L["Tile the background texture"],
						},
						TileSize = {
							type = "range",
							order = 60,
							name = L["Background Tile Size"],
							desc = L["The size used to tile the background texture"],
							min = 16, max = 256, step = 1,
							disabled = function() return not db.Background.Tile end,
						},
						EdgeSize = {
							type = "range",
							order = 70,
							name = L["Border Thickness"],
							desc = L["The thickness of the border"],
							min = 1, max = 24, step = 1,
						},
					},
				},
			},
		},
		show = {
			type = 'group',
			name = L["Show When"],
			desc = L["Set rules for when to show Utopia"],
			get = function(info) return db.show[info[#info]] end,
			set = function(info, value)
				db.show[info[#info]] = value
				module:ShowWhen()
			end,
			order = 50,
			args = {
				always = {
					type = 'toggle',
					name = L["Always Show"],
					desc = L["Always show the bar"],
					order = 1,
				},
				solo = {
					type = 'toggle',
					name = SOLO,
					desc = L["Show when solo"],
					disabled = function() return db.show.always end,
					order = 10,
				},
				party = {
					type = 'toggle',
					name = PARTY,
					desc = L["Show when in a party"],
					disabled = function() return db.show.always end,
					order = 20,
				},
				raid = {
					type = 'toggle',
					name = RAID,
					desc = L["Show when in a raid"],
					disabled = function() return db.show.always end,
					order = 30,
				},
				enemy = {
					type = 'toggle',
					name = L["Enemy"],
					desc = L["Only show when targetting a hostile unit"],
					disabled = function() return db.show.always end,
					order = 35,
				},
				boss = {
					type = 'toggle',
					name = L["Bosses"],
					desc = L["Only show when targetting boss mobs"],
					disabled = function() return db.show.always or db.show.enemy end,
					order = 40,
				},
			},
		},
	},
}

AceConfig:RegisterOptionsTable("Utopia", module.options)

-- FirstInitData
function module:FirstInitData()
	local function DoList(list)
		for name,info in pairs(list) do
			local newlist = {}
			for spell,spellInfo in pairs(info.spells) do
				if (type(spell) == "number") then
					local id = spell
					spell = GetSpellInfo(id)
					if (not spell) then
						error("Missing spell for ID %d in %s", id, name)
					end
					spellInfo.id = id
					newlist[spell] = spellInfo
					info.spells[id] = nil
				else
					newlist[spell] = spellInfo
					info.spells[spell] = nil
				end
			end
			info.spells = newlist
		end
	end
	DoList(self.buffs)
	DoList(self.debuffs)

	local newList = {}
	for id,info in pairs(Utopia.otherBuffs) do
		newList[GetSpellInfo(id)] = info
		info.id = id
	end
	Utopia.otherBuffs = newList

	self.FirstInitData = nil
end

-- AddExclusiveData
function module:AddExclusiveData(spellInfo)
	local ex = spellInfo.exclusive or spellInfo.totem
	if (ex) then
		local tab = self.exclusive[ex]
		if (not tab) then
			tab = new()
			self.exclusive[ex] = tab
		end
		if (not tab[spellInfo.name]) then
			tab[spellInfo.name] = true
		end
	end
end

-- InitData
function module:InitData()
	if (self.lookup) then
		del(self.lookup.buffs)
		del(self.lookup.debuffs)
		del(self.lookup)
	end
	deepDel(self.classes)
	deepDel(self.exclusive)

	local menu = module.options.args.icons.args
	menu.buffs.args = deepDel(menu.buffs.args)
	menu.debuffs.args = deepDel(menu.debuffs.args)

	self.lookup = new()
	self.lookup.buffs = new()
	self.lookup.debuffs = new()
	self.classes = new()
	self.exclusive = new()					-- Exclusives lookup

	menu.buffs.args = new()
	menu.debuffs.args = new()

	if (self.FirstInitData) then
		self:FirstInitData()
	end

	local list = new()
	for name,info in pairs(self.buffs) do
		tinsert(list, name)
		info.key = name
		for spell,spellInfo in pairs(info.spells) do
			spellInfo.name = spell
			spellInfo.key = name
			self.lookup.buffs[spell] = spellInfo
			if (spellInfo.alternate) then
				self.lookup.buffs[spellInfo.alternate] = spellInfo
			end

			self:AddExclusiveData(spellInfo)

			local c = self.classes[spellInfo.class]
			if (not c) then
				c = new()
				self.classes[spellInfo.class] = c
			end
			if (not c.buffs) then
				c.buffs = new()
			end
			c.buffs[name] = spell
		end
	end
	sort(list)
	for i,name in ipairs(list) do
		local m = new()
		menu.buffs.args[tostring(i)] = m
		m.type = "toggle"
		m.name = name
		m.desc = name
		m.order = i
	end
	del(list)

	local list = new()
	for name,info in pairs(self.debuffs) do
		tinsert(list, name)
		info.key = name
		for spell,spellInfo in pairs(info.spells) do
			spellInfo.name = spell
			spellInfo.key = name
			self.lookup.debuffs[spell] = spellInfo

			self:AddExclusiveData(spellInfo)

			local c = self.classes[spellInfo.class]
			if (not c) then
				c = new()
				self.classes[spellInfo.class] = c
			end
			if (not c.debuffs) then
				c.debuffs = new()
			end
			c.debuffs[name] = spell
		end
	end
	sort(list)
	for i,name in ipairs(list) do
		local m = new()
		menu.debuffs.args[tostring(i)] = m
		m.type = "toggle"
		m.name = name
		m.desc = name
		m.order = i
	end
	del(list)

	del(self.mappingLookup)
	self.mappingLookup = new()

	for targetSpell,data in pairs(self.mapping) do
		data.key = targetSpell
		for sourceSpell,info in pairs(data.list) do
			-- Mappings now a list of mappings. Some spells have more than 1 possible
			-- ie: Devotion Aura -> Improved Devotion Aura OR to Swift Retribution OR to Sanctified Retribution
			local cur = self.mappingLookup[sourceSpell]
			if (not cur) then
				cur = new()
				self.mappingLookup[sourceSpell] = cur
			end
			tinsert(cur, data)

			--self.mappingLookup[sourceSpell] = data
		end
	end
end

-- SavePosition
function module:SavePosition(frame)
	local point, _, relPoint, x, y = frame:GetPoint(1)
	if (point) then
		if (not self.db.char.pos) then
			self.db.char.pos = {}
		end
		self.db.char.pos[frame:GetName()] = {point, "UIParent", relPoint, x, y}
	end
end

-- RestorePosition
function module:RestorePosition(frame)
	if (type(self.db.char.pos) == "table") then
		-- Convert old pos to new pos savedvars
		if (#self.db.char.pos == 5 and not self.db.char.pos.UtopiaAnchorFrame) then
			local old = self.db.char.pos
			self.db.char.pos = new()
			self.db.char.pos.UtopiaAnchorFrame = old
		end
	end

	local pos = self.db.char.pos and self.db.char.pos[frame:GetName()]
	if (pos) then
		frame:SetPoint(unpack(pos))
	else
		frame:SetPoint("CENTER", UIParent, "TOP", 0, -150)
	end
end

-- myLevelRank
local function myLevelRank(ranks)
	if (ranks) then
		local lvl = UnitLevel("player")
		local rank
		for i,level in pairs(ranks) do
			if (lvl >= level) then
				rank = i
			end
		end
		return rank
	end
end

-- GetAmount
function module:GetAmount(info, short, getMaximum, auraCaster, auraName, auraRank, auraTex, auraCount, auraDebuffType, auraDuration, auraEndTime, auraIsStealable)
--@debug@
-- So I can run from command line :)
	if (type(info) == "string") then
		info = self.lookup.buffs[info] or self.lookup.debuffs[info]
	end
--@no-debug@
	local casterName = auraCaster and UnitFullName(auraCaster) or auraCaster
	if (not casterName and auraName) then
		-- We've been given buff info of someone out of sight, so caster name is missing, so we need to use UNKNOWN
		-- So that we don't pull in the talented max amount by mistake
		casterName = UNKNOWN
	end
	local ret = info.amount
	local suffix

	if (info.minLevel and UnitLevel("player") < info.minLevel) then
		return "--"
	end

	local name, rank, stacks
	if (auraName) then
		rank = auraRank

		if (rank and auraName == info.alternate and info.alternateOffset) then
			rank = rank + info.alternateOffset
		end
		if (info.maxStacks) then
			stacks = auraCount
		elseif (auraCount and auraCount > 1) then
			error(format("%s has %d stacks. info.maxStacks not set", info.name, auraCount))
		end
	else
		rank = myLevelRank(info.rankLevels)
		if (not rank and info.rankLevels) then
			rank = -1
		end
	end

	if (info.amountFunc) then
		ret = info.amountFunc(casterName)
	elseif (info.amountsPerStack) then
		ret = info.amountsPerStack[rank or #info.amountsPerStack]
		if (not ret) then
			error(format("No info.amountsPerStack[%s] for %s", tostring(rank or #info.amountsPerStack), tostring(info.name)))
		end
		if (not short) then
			suffix = format(L["(for %d %s)"], stacks or info.maxStacks, L["stacks"])
		end
	elseif (info.amountPerStack) then
		ret = info.amountPerStack
		assert(ret)
		if (not short) then
			suffix = format(L["(for %d %s)"], stacks or info.maxStacks, L["stacks"])
		end
	elseif (info.amounts) then
		if (rank == -1) then
			return "--"
		else
			ret = info.amounts[rank or #info.amounts]
			if (not ret) then
				error(format("No info.amounts[%s] for %s", tostring(rank or #info.amounts), tostring(info.name)))
			end
		end
	end

	if (info.improvedAmountPerTalentPoint) then
		if (casterName) then
			local points = self:HasTalent(casterName, info.improved)
			if (points) then
				if (points > info.maxTalentPoints) then
					self:RefreshTalents(casterName)
					points = 0
--@debug@
					self:Print(format("Resetting %s's talent points due to mismatch with maxTalentPoints for %s", casterName, info.name))
--@no-debug@
				end
				ret = ret + info.improvedAmountPerTalentPoint * points
			end
		else
			local potentialLevel = self.spellPotential[info.name] or (db.defaultMax and info.maxTalentPoints or 0)
			if (not getMaximum and potentialLevel) then
				if (potentialLevel > 0) then
					ret = ret + info.improvedAmountPerTalentPoint * potentialLevel
				end
			else
				ret = ret + info.improvedAmountPerTalentPoint * (info.maxTalentPoints or 1)
			end
		end
		ret = ret and floor(ret)

	elseif (info.improvedPercentagePerTalentPoint) then
		if (casterName) then
			local points = self:HasTalent(casterName, info.improved)
			if (points) then
				if (points > info.maxTalentPoints) then
					self:RefreshTalents(casterName)
					points = 0
--@debug@
					self:Print(format("Resetting %s's talent points due to mismatch with maxTalentPoints for %s", casterName, info.name))
--@no-debug@
				end
				if (info.amountType == "%") then
					ret = ret + info.improvedPercentagePerTalentPoint * points
				else
					ret = ret * (1 + (info.improvedPercentagePerTalentPoint * points / 100))
				end
			end
		else
			local potentialLevel = self.spellPotential[info.name] or (db.defaultMax and info.maxTalentPoints or 0)
			if (not getMaximum and potentialLevel) then
				if (potentialLevel > 0) then
					if (info.amountType == "%") then
						ret = ret + info.improvedPercentagePerTalentPoint * potentialLevel
					else
						ret = ret * (1 + (info.improvedPercentagePerTalentPoint * potentialLevel / 100))
					end
				end
			else
				if (info.amountType == "%") then
					ret = ret + info.improvedPercentagePerTalentPoint * (info.maxTalentPoints or 1)
				else
					ret = ret * (1 + (info.improvedPercentagePerTalentPoint * (info.maxTalentPoints or 1) / 100))
				end
			end
		end
		ret = ret and floor(ret)

	elseif (info.amountPerTalentPoint) then
		if (casterName) then
			local points = self:HasTalent(casterName, info.requiredTalent)
			if (points) then
				if (points > info.maxTalentPoints) then
					self:RefreshTalents(casterName)
					points = info.maxTalentPoints
--@debug@
					self:Print(format("Resetting %s's talent points due to mismatch with data", casterName))
--@no-debug@
				end
				ret = info.amountPerTalentPoint * points
			end
		else
			local potentialLevel = self.spellPotential[info.name]
			if (not getMaximum and potentialLevel) then
				if (potentialLevel > 0) then
					ret = info.amountPerTalentPoint * potentialLevel
				end
			else
				ret = info.amountPerTalentPoint * (info.maxTalentPoints or 1)
			end
		end
		ret = ret and floor(0.5 + ret)
	end

	if (ret and info.maxStacks) then
		if (info.assumeMaxStacks) then
			ret = ret * info.maxStacks
		else
			ret = ret * (stacks or info.maxStacks or 1)
		end
		if (not short) then
			suffix = format(L["(for %d %s)"], (info.assumeMaxStacks and info.maxStacks) or stacks or info.maxStacks, info.stacksDescription or L["stacks"])
		end
--@debug@
	else
		if (stacks and stacks > 1 and not info.maxStacks) then
			error(format("%s has %d stacks. info.maxStacks not set", info.name, stacks))
		end
--@no-debug@
	end

	if (ret) then
		--ret = floor(0.5 + ret)
		if (suffix) then
			return format("%s%s %s", ret, info.amountType or "", suffix), ret
		else
			return format("%s%s", ret, info.amountType or ""), ret
		end
	end
	return ""
end

-- OnEnter
function module:OnEnter()
	if (not self.popup) then
		self:MakePopupButton()
	end
	if (not self.popup:IsShown()) then
		self.popup:ClearAllPoints()
		if (db.popupPos == "top") then
			self.popup:SetPoint("BOTTOMRIGHT", self.frame, "TOPRIGHT", 0, db.Background.EdgeSize / 2)
		else
			self.popup:SetPoint("TOPRIGHT", self.frame, "BOTTOMRIGHT", 0, -db.Background.EdgeSize / 2)
		end

		if (not self.popup:IsShown()) then
			self.popup:Show()
		end
		if (self.popup.fadeOut:IsPlaying()) then
			self.popup.fadeOut:Stop()
		end
		if (not self.popup.fadeIn:IsPlaying()) then
			self.popup.fadeIn:Play()
		end
	end
end

-- OnLeave
function module:OnLeave()
	if (self.popup) then
		if (not MouseIsOver(self.frame) and not MouseIsOver(self.popup)) then
			self.popup.fadeIn:Stop()
			if (not self.popup.fadeOut:IsPlaying()) then
				self.popup.fadeOut:Play()
			end
		end
	end
end

local function dragStart()
	if (module.popup) then
		module.popup.fadeIn:Stop()
		module.popup.fadeOut:Stop()
		module.popup:Hide()
	end
	module.frame:StartMoving()
	GameTooltip:Hide()
end
local function dragStop()
	module.frame:StopMovingOrSizing()

	local point, parent, relpoint, x, y = module.frame:GetPoint(1)
	parent = UIParent
	local leave
	if (db.popupPos == "top") then
		if (point:find("BOTTOM")) then
			y = y + module.frame:GetHeight()
			point = point:gsub("BOTTOM", "TOP")
		elseif (not point:find("TOP")) then
			y = y + module.frame:GetHeight() / 2
			if (point == "CENTER") then
				point = "TOP"
			else
				point = "TOP"..point
			end
		else
			leave = true
		end
	else
		if (point:find("TOP")) then
			y = y - module.frame:GetHeight()
			point = point:gsub("TOP", "BOTTOM")
		elseif (not point:find("BOTTOM")) then
			y = y - module.frame:GetHeight() / 2
			if (point == "CENTER") then
				point = "BOTTOM"
			else
				point = "BOTTOM"..point
			end
		else
			leave = true
		end
	end
	if (not leave) then
		module.frame:ClearAllPoints()
		module.frame:SetPoint(point, UIParent, relpoint, x, y)
	end

	module:SavePosition(module.frame)
end

local info
local function sortList(a, b)
	local aa = info.spells[a]
	local bb = info.spells[b]
	aa = module:GetAmount(aa, true)
	bb = module:GetAmount(bb, true)

	if (aa == bb) then
		if (type(a) == type(b)) then
			return a < b
		else
			return (tonumber(a) or 0) < (tonumber(b) or 0)
		end
	else
		if (type(aa) == type(bb)) then
			return aa > bb
		else
			return (tonumber(aa) or 0) > (tonumber(bb) or 0)
		end
	end
end

-- MakePopupButton
function module:MakePopupButton()
	local fname = self.frame:GetName().."Popup"
	local f = CreateFrame("Frame", fname, self.frame)
	self.popup = f
	f:SetScale(0.5)
	f:SetHeight(36)
	f:SetWidth(144)

	f:SetScript("OnShow",
		function(self)
			self:SetAlpha(0)
		end
	)

	local plusIcon = select(3, GetSpellInfo(28062))
	local minusIcon = select(3, GetSpellInfo(28085))

	local b = CreateFrame("Button", fname.."Button1", self.popup, "ActionButtonTemplate")
	b:SetPoint("TOPRIGHT")
	b.icon = getglobal(fname.."Button1Icon")
	b.icon:SetTexture(self.frame.mode == "buffs" and plusIcon or minusIcon)

	b:SetScript("OnClick",
		function(self)
			module.frame:SetMode(module.frame.mode == "buffs" and "debuffs" or "buffs")
			self.icon:SetTexture(module.frame.mode == "buffs" and plusIcon or minusIcon)
		end
	)
	b:SetScript("OnEnter",
		function(self)
			module.popup.fadeOut:Stop()
			if (not module.popup:IsShown() or module.popup:GetAlpha() < 1) then
				if (not module.popup.fadeIn:IsPlaying()) then
					module.popup.fadeIn:Play()
				end
			end

			GameTooltip:SetOwner(self, "LEFT")
			GameTooltip:SetText(L["Mode"], 1, 1, 1)
			GameTooltip:AddLine(format(L["Switch mode between |T%s:0|t (Buffs) and |T%s:0|t (Debuffs)"], plusIcon, minusIcon))
			GameTooltip:Show()
		end
	)
	b:SetScript("OnLeave",
		function(self)
			GameTooltip:Hide()
			module:OnLeave()
		end
	)

	local manualIcon = "Interface\\Addons\\Utopia\\Textures\\filtered-off"
	local autoIcon = "Interface\\Addons\\Utopia\\Textures\\filtered-on"

	local b2 = CreateFrame("Button", fname.."Button2", self.popup, "ActionButtonTemplate")
	b2:SetPoint("TOPRIGHT", b, "TOPLEFT")
	b2.icon = getglobal(fname.."Button2Icon")
	b2.icon:SetTexture(db.automatic and autoIcon or manualIcon)

	b2:SetScript("OnClick",
		function(self)
			db.automatic = not db.automatic
			module.frame.lastTargetGUID = nil
			module.frame:SetMode()
			self.icon:SetTexture(db.automatic and autoIcon or manualIcon)
		end
	)
	b2:SetScript("OnEnter",
		function(self)
			module.popup.fadeOut:Stop()
			if (not module.popup:IsShown() or module.popup:GetAlpha() < 1) then
				if (not module.popup.fadeIn:IsPlaying()) then
					module.popup.fadeIn:Play()
				end
			end

			GameTooltip:SetOwner(self, "LEFT")
			GameTooltip:SetText(L["Automatic Toggle"], 1, 1, 1)
			GameTooltip:AddLine(format(L["Switch mode between |T%s:0|t (Automatic Icon Selection) and |T%s:0|t (Manual Icons)"], autoIcon, manualIcon), nil, nil, nil, 1)
			GameTooltip:Show()
		end
	)
	b2:SetScript("OnLeave",
		function(self)
			GameTooltip:Hide()
			module:OnLeave()
		end
	)

	local allIcon = "Interface\\Addons\\Utopia\\Textures\\automatic-on"
	local availableIcon = "Interface\\Addons\\Utopia\\Textures\\automatic-off"

	local b3 = CreateFrame("Button", fname.."Button3", self.popup, "ActionButtonTemplate")
	b3:SetPoint("TOPRIGHT", b2, "TOPLEFT")
	b3.icon = getglobal(fname.."Button3Icon")
	b3.icon:SetTexture(db.onlyAvailable and allIcon or availableIcon)

	b3:SetScript("OnClick",
		function(self)
			db.onlyAvailable = not db.onlyAvailable
			module.frame.lastTargetGUID = nil
			module.frame:SetMode()
			self.icon:SetTexture(db.onlyAvailable and allIcon or availableIcon)
		end
	)
	b3:SetScript("OnEnter",
		function(self)
			module.popup.fadeOut:Stop()
			if (not module.popup:IsShown() or module.popup:GetAlpha() < 1) then
				if (not module.popup.fadeIn:IsPlaying()) then
					module.popup.fadeIn:Play()
				end
			end

			GameTooltip:SetOwner(self, "LEFT")
			GameTooltip:SetText(L["Available Only Toggle"], 1, 1, 1)
			GameTooltip:AddLine(format(L["Switch mode between |T%s:0|t (All Icons) and |T%s:0|t (Only Available Icons)"], allIcon, availableIcon), nil, nil, nil, 1)
			GameTooltip:Show()
		end
	)
	b3:SetScript("OnLeave",
		function(self)
			GameTooltip:Hide()
			module:OnLeave()
		end
	)

	f.fadeIn = f:CreateAnimationGroup()
	local a = f.fadeIn:CreateAnimation("Alpha")
	f.fadeIn.alpha = a
	a:SetDuration(0.5)
	a:SetChange(1)
	a:Play()
	a:Stop()
	a:SetScript("OnUpdate",
		function(self)
			self:GetRegionParent():SetAlpha(self:GetProgress())
		end
	)

	f.fadeOut = f:CreateAnimationGroup()
	a = f.fadeOut:CreateAnimation("Alpha")
	f.fadeOut.alpha = a
	a:SetDuration(0.5)
	a:SetChange(-1)
	a:SetStartDelay(0.5)
	a:SetScript("OnFinished",
		function(self)
			module.popup:Hide()
		end
	)
	a:SetScript("OnUpdate",
		function(self)
			self:GetRegionParent():SetAlpha(1 - self:GetProgress())
		end
	)

	f:SetAlpha(0)
	f:Hide()
	return f
end

-- MakeBaseFrame
function module:MakeBaseFrame(name, unit)
	local frame = CreateFrame("Frame", name, UIParent)
	frame:SetHeight(1)
	frame:SetWidth(1)
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)

	module:SetScale()
	module:RestorePosition(frame)

	frame.icons = {}

	return frame
end

-- EnableBar
function module:EnableBar(_, n)
	db.bar = n
	if (db.bar) then
		if (module.CreateFramesBar) then
			module:CreateFramesBar()
		end
		self.frame.bar:Show()
	else
		if (self.frame.bar) then
			self.frame.bar:Hide()
		end
	end
	self:SetFrameSize()
end

-- CreateFramesBar
function module:CreateFramesBar()
	if (db.bar) then
		local frame = self.frame
		local bar = CreateFrame("StatusBar", frame:GetName().."Bar", frame)
		frame.bar = bar

		bar:SetStatusBarTexture(SM:Fetch("statusbar", db.texture))
		bar:SetStatusBarColor(0, 1, 0)
		bar:SetMinMaxValues(0, frame.shown or 1)

		bar:RegisterForDrag("LeftButton")
		bar:SetScript("OnDragStart", dragStart)
		bar:SetScript("OnDragStop", dragStop)
		bar:SetScript("OnEnter", function() module:OnEnter() end)
		bar:SetScript("OnLeave", function() module:OnLeave() end)

		bar.bg = frame.bar:CreateTexture(nil, "BACKGROUND")
		bar.bg:SetAllPoints()
		bar.bg:SetTexture(SM:Fetch("statusbar", db.texture))
		bar.bg:SetVertexColor(0, 1, 0, 0.3)

		if (not db.locked) then
			bar:EnableMouse(true)
		end
		self.CreateFramesBar = nil
	end
end

-- onEnter
local function onEnter(self)
	module:OnEnter()
	GameTooltip:SetOwner(self, "LEFT")
	GameTooltip:SetText(self.key, 1, 1, 1)

	info = module[self.mode][self.key]					-- DO NOT LOCALIZE, it's up there with the sortList func
	if (not info) then
		return
	end

	local list = new()
	for name,spellInfo in pairs(info.spells) do
		tinsert(list, name)
	end
	sort(list, sortList)

	local active, missing = module:GetActiveAurasForKey(self:GetParent().unit, self.mode, self.key)

	for i = 1,#list do
		local name = list[i]
		local spellInfo = info.spells[name]

		local canDo = module.spellPotential[name]
		if (not canDo) then
			if (petClasses[spellInfo.class]) then
				canDo = module.petPotential and module.petPotential[name]
			end
		end
		local div = canDo and 1 or 2

		local c = RAID_CLASS_COLORS[spellInfo.class]
		local potentialAmount = module:GetAmount(spellInfo)

		if (spellInfo.pet) then
			GameTooltip:AddDoubleLine(format("%s (%s)", spellInfo.source or name, spellInfo.pet), potentialAmount, c.r / div, c.g / div, c.b / div, 1 / div, 1 / div, 1 / div)
		else
			if (spellInfo.source) then
				GameTooltip:AddDoubleLine(format("%s|cFF808080 (|r%s|cFF808080)", spellInfo.source, name), potentialAmount, c.r / div, c.g / div, c.b / div, 1 / div, 1 / div, 1 / div)
			else
				GameTooltip:AddDoubleLine(spellInfo.source or name, potentialAmount, c.r / div, c.g / div, c.b / div, 1 / div, 1 / div, 1 / div)
			end
		end

		if (canDo or (active and active[name])) then
  			GameTooltip:AddTexture(active and active[name] and "Interface\\RAIDFRAME\\ReadyCheck-Ready" or "Interface\\RaidFrame\\ReadyCheck-NotReady")
		end

		local act = active and active[name]
		if (act) then
			local auraCaster = act.caster
			local pet
			if (auraCaster and strfind(auraCaster, "pet")) then
				pet = auraCaster
				auraCaster = auraCaster:gsub("(%a+)pet(%d+)", "%1%2")
			end

			local actualAmount = module:GetAmount(spellInfo, nil, nil, auraCaster, act.name, act.rank, act.tex, act.count, act.debuffType, act.duration, act.endTime, act.isStealable)
			local who
			if (pet) then
				who = format(L["  by %s (%s's pet)"], module:ColourPlayer(pet), module:ColourPlayer(auraCaster))
			else
				who = format(L["  by %s"], auraCaster and module:ColourPlayer(auraCaster) or UNKNOWN)
			end

			if (actualAmount < potentialAmount) then
				GameTooltip:AddDoubleLine(who, actualAmount, 0.5, 0.5, 0.5, 1, 0.25, 0.25)
			else
				GameTooltip:AddLine(who, 0.5, 0.5, 0.5)
			end
		end

		if (spellInfo.id == Utopia.bloodlustID) then
			local ready, cooldown, whoCanCast, benefitRatio = module:GetBloodlustAvailable()
			if (ready or (benefitRatio and benefitRatio > 1)) then
				GameTooltip:AddLine(format(L["|cFF00FF00Ready!|r (%s)"], whoCanCast), 1, 1, 1)
			elseif (cooldown and cooldown ~= -1) then
				GameTooltip:AddLine(format(L["|cFFFF8080Cooldown:|r %s"], date("%X", cooldown)), 1, 1, 1)
			end
			if (benefitRatio and benefitRatio > 0 and benefitRatio < 100) then
				GameTooltip:AddLine(format(L["%d%% without Sated debuff"], benefitRatio), 1, 1, 1)
			end
		end
	end
	del(list)
	GameTooltip:Show()

	deepDel(active)
	del(missing)
end

local function onLeave(self)
	module:OnLeave()
	GameTooltip:Hide()
end

local function iconDragStart(self, button)
--d("iconDragStart(%s, %s)", tostring(self:GetName()), tostring(button))
	if (button == "RightButton") then
		
		
	else
		dragStart(self, button)
	end
end
local function iconDragStop(self, button)
--p("iconDragStop(%s, %s)", tostring(self:GetName()), tostring(button))
	if (button == "RightButton") then
		
	else
		dragStop(self, button)
	end
end

local function onUpdateRosterTrigger(self, elapsed)
	module:OnRaidRosterUpdate()
	self:SetScript("OnUpdate", nil)
end

-- InitFramesDebuff
function module:InitFrames()
	local frame = self:MakeBaseFrame("UtopiaAnchorFrame", "target")
	self.frame = frame

	local function iconActivate(self, debuff, caster)
		self.toggle = "on"
		local c = db.colour.active

		local info = module.lookup[self.mode][debuff]
		if (info and info.improved) then
			local points = module:HasTalent(caster, info.improved)
			if (not points or points < info.maxTalentPoints) then
				c = db.colour.partActive
			end
		end

		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)

		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function iconShowCount(self, count)
		if (count and count < 100) then
			if (not self.count) then
				self.count = self:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
				self.count:SetPoint("TOPLEFT", -2, 0)
				self.count:SetPoint("BOTTOMRIGHT", 2, 0)
			end
			self.count:SetFormattedText("%d%%", count)
			local r, g, b = SmoothColour(count / 100)
			self.count:SetTextColor(r, g, b)
			self.count:Show()

		elseif (self.count) then
			self.count:Hide()
		end

		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function iconDeactivate(self, tempOnly)
		self.toggle = "off"
		local c = db.colour[tempOnly and "temporary" or "inactive"]
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)
		if (self.count) then
			self.count:Hide()
		end
		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function iconUnavailable(self, excluded)
		self.toggle = nil
		if (self:IsShown()) then
			module.frame.totalAvailable = max(0, module.frame.totalAvailable - 1)
		end
		local c = db.colour[excluded and "excluded" or "unavailable"]
		self.tex:SetVertexColor(c.r, c.g, c.b, c.a)
		if (self.count) then
			self.count:Hide()
		end
		if (GameTooltip:IsOwned(self)) then
			onEnter(self)
		end
	end

	local function buttonOnClick(self, button)
		if (button == "LeftButton") then
			module:ToggleDetailsSuper(self.mode, self.key)
		elseif (button == "RightButton") then
			module:ReportMissingIcon(self:GetParent().unit, self.mode, self.key)
		end
	end

	function frame:SetMode(mode)
		if (mode) then
			assert(mode == "buffs" or mode == "debuffs")
			db.mode = mode
		elseif (not db.mode) then
			db.mode = "debuffs"
		end
		mode = db.mode
		self.mode = mode

		del(self.keys)
		del(self.keysLookup)
		self.keys = new()
		self.keysLookup = new()
		for key,data in pairs(module[mode]) do
			tinsert(self.keys, key)
		end
		sort(self.keys)
		for i,key in ipairs(self.keys) do
			self.keysLookup[key] = i
		end

		frame.shown = 0
		for i = 1,#self.keys do
			local icon = self.icons[i]
			if (not icon) then
				-- TODO: Mouseover highlight and other eye-candy
				-- And fix the sizing problem on startup
				icon = CreateFrame("Button", self:GetName()..i, frame)
				icon.tex = icon:CreateTexture(nil, "OVERLAY")
				icon.tex:SetAllPoints()

				icon.parent = self
				icon:SetWidth(db.size)
				icon:SetHeight(db.size)
				icon:EnableMouse(true)

				if (not db.locked) then
					icon:RegisterForDrag("LeftButton")		-- ,RightButton
				end
				icon:RegisterForClicks("AnyUp")

				icon:SetScript("OnEnter", onEnter)
				icon:SetScript("OnLeave", onLeave)
				icon:SetScript("OnDragStart", iconDragStart)
				icon:SetScript("OnDragStop", iconDragStop)
				icon:SetScript("OnClick", buttonOnClick)

				icon.Activate = iconActivate
				icon.Deactivate = iconDeactivate
				icon.Unavailable = iconUnavailable
				icon.ShowCount = iconShowCount

				self.icons[i] = icon
			end

			icon.mode = mode
			icon.key = self.keys[i]
			local data = module[mode][icon.key]
			icon.tex:SetTexture(data.icon)

			if (db.automatic) then
				local mana = UnitPowerMax("player", 0) > 0
				local role = data.interested and module:MyRole()
				if (role) then
					if (strfind(data.interested, role)) then
						-- It's something interesting to my class, so let's see it
						icon:Show()

					elseif (mana and strfind(data.interested, "manauser")) then
						icon:Show()

					else
						-- Might not be something interesting, but if we can cast it then we show it
						local canCast
						for spellName,spellInfo in pairs(data.spells) do
							if (spellInfo.class == playerClass) then
								if (spellInfo.requiredTalent) then
									if (module:HasTalent(playerName, spellInfo.requiredTalent)) then
										canCast = true
										break
									end
								else
									canCast = true
									break
								end
							end
						end
						if (canCast) then
							icon:Show()
						else
							icon:Hide()
						end
					end
				else
					icon:Show()
				end
			else
				if (db.disabled and db.disabled[mode] and db.disabled[mode][icon.key]) then
					icon:Hide()
				else
					icon:Show()
				end
			end
			if (icon:IsShown()) then
				frame.shown = frame.shown + 1
			end
		end
		for i = #self.keys + 1,#self.icons do
			local icon = self.icons[i]
			if (icon) then
				icon:Hide()
			end
		end
		if (self.bar) then
			self.bar:SetMinMaxValues(0, frame.shown)
		end

		self:PLAYER_TARGET_CHANGED()
		
		module:SetOrientation()
	end

	self:CreateFramesBar()

	frame:SetScript("OnUpdate",
		function(self, elapsed)
			if (module.rosterUpdate) then
				module.rosterUpdate = nil
				module:OnRaidRosterUpdate()
			end
		end
	)
	frame:SetScript("OnShow",
		function(self)
			self:UpdateBuffs()
			self:RegisterEvent("UNIT_AURA")
			self:RegisterEvent("PLAYER_TARGET_CHANGED")
		end
	)
	frame:SetScript("OnHide",
		function(self)
			self:UnregisterEvent("UNIT_AURA")
			self:UnregisterEvent("PLAYER_TARGET_CHANGED")
			self:UnregisterEvent("PLAYER_FOCUS_CHANGED")
		end
	)
	frame:SetScript("OnEvent",
		function(self, event, ...)
			local f = self[event]
			if (f) then
				f(self, ...)
			end
		end
	)

	-- UNIT_AURA
	function frame:UNIT_AURA(unit)
		if (unit == self.unit) then		-- Don't use UnitIsUnit(), because we only want to do this once. We'll get the event for each immediate unit anyway
			self:UpdateBuffs()
		end

		if (unit == "player") then
			if (playerClass == "DRUID" or playerClass == "DEATHKNIGHT") then
				module:CheckRoleChange()
			end
		else
			if (not UnitIsVisible(unit)) then
				if (not self.outOfSight) then
					self.outOfSight = new()
				end
				self.outOfSight[UnitFullName(unit)] = true
			end
		end
	end

	-- PLAYER_FOCUS_CHANGED
	function frame:PLAYER_FOCUS_CHANGED()
		if (self.unit == "focus") then
			self:PLAYER_TARGET_CHANGED()
		end
	end

	-- PLAYER_TARGET_CHANGED
	function frame:PLAYER_TARGET_CHANGED()
		if (self.mode == "buffs") then
			if (db.allPlayers and (UnitInRaid("target") or UnitInParty("target"))) then
				self.unit = "target"
			else
				self.unit = "player"
			end
		else
			if (UnitIsFriend("player", "target")) then
				if (not UnitIsFriend("player", "focus")) then
					self.unit = "focus"
				elseif (not UnitIsFriend("player", "targettarget")) then
					self.unit = "targettarget"
				else
					self.unit = "target"
				end
			else
				self.unit = "target"
			end
		end
		local guid = UnitGUID(self.unit)
		if (guid ~= self.lastTargetGUID) then
			self:UpdateBuffs()
			self.lastTargetGUID = guid
		end
		--module:ShowWhen()
		self:SetTitle()
	end

	-- AllOff
	function frame:AllOff()
		local reorient
		self.totalAvailable = 0
		self.totalAvailableNonTemporary = 0
		for i,icon in ipairs(self.icons) do
			local section = module[self.mode][icon.key]
			local exclude = section and section.exclude
			local tempOnly

			if (icon:IsShown()) then
				self.totalAvailable = self.totalAvailable + 1

				local temp, nontemp = 0, 0
				for spellName,info in pairs(section.spells) do
					if (module.spellPotential[spellName]) then
						if (info.temporary or (info.totem and db.totemsTemporary)) then
							temp = temp + 1
						else
							nontemp = nontemp + 1
						end
					end
				end

				if (nontemp > 0 and temp == 0) then
					self.totalAvailableNonTemporary = self.totalAvailableNonTemporary + 1
				elseif (temp > 0 and nontemp == 0) then
					tempOnly = true
				end
			end
			icon.data = deepDel(icon.data)

			if (exclude and exclude(self.unit)) then
				icon:Unavailable()
			else
				local pot = module.potential[self.mode]
				if (pot and pot[icon.key]) then
					icon:Deactivate(not InCombatLockdown() and tempOnly)
				else
					if (module.petPotentialKeys and module.petPotentialKeys[icon.key]) then
						icon:Deactivate(not InCombatLockdown() and tempOnly)
					else
						if (db.onlyAvailable) then
							if (icon:IsShown()) then
								icon:Hide()
								reorient = true
							end
						else
							icon:Unavailable()
						end
					end
				end
			end
		end
		if (reorient) then
			module:SetOrientation()
		end
		if (self.bar) then
			self.bar:SetMinMaxValues(0, InCombatLockdown() and self.totalAvailable or self.totalAvailableNonTemporary)
			self.bar:SetValue(0)
			self.bar:SetStatusBarColor(1, 0, 0)
			self.bar.bg:SetVertexColor(1, 0, 0, 0.3)
		end
	end

	-- GetIconForKey
	function frame:GetIconForKey(key)
		local index = self.keysLookup[key]
		if (index) then
			return self.icons[index]
		end
	end

	-- GetIconForSpell
	function frame:GetIconForSpell(spellName)
		local spell = module.lookup[self.mode][spellName]
		if (not spell) then
			return
		end
		return self:GetIconForKey(spell.key), spell
	end

	-- AddActiveAura
	-- *** RECURSIVE ***
	function frame:AddActiveAura(name, caster, count)
		local on = 0
		local icon, spell = self:GetIconForSpell(name)
		if (icon and icon:IsShown()) then
			icon:Activate(name, caster)
			if (count and count > 0 and spell.maxStacks) then
				icon:ShowCount(100 * count / spell.maxStacks)
			end
			on = on + 1

			if (not icon.data) then
				icon.data = new()
			end
			icon.data[name] = caster
		end

		local exclusive
		if (spell) then
			exclusive = spell.exclusive or spell.totem
		else
			-- Check if it's in the exclusive list that lists un-monitored spells
			local ex = module.otherBuffs[name]
			if (ex) then
				exclusive = ex.exclusive or ex.totem
			end
		end

		if (exclusive) then
			-- Handle exclusive auras (Paladin Judgements/Blessings, Warlock Curses, Warrior Shouts, Totems etc)

			-- We have a pre-built list of paired exclusive spell names in module.exclusive
			local list = module.exclusive[exclusive]
			if (list) then
				for spellname in pairs(list) do
					if (spellname ~= name) then			-- Don't check current spell
						local icon2, spell2 = self:GetIconForSpell(spellname)
						if (icon2 and icon2:IsShown() and icon2.toggle == "off") then
							-- Found a potentially unavailable aura because the caster has already done something
							-- of the same class. Now we need to check if there are any other people who could cast
							-- this before we flag the icon as unavailable

							-- Now we have a possibly conflicting icon, we'll go through each of the spells that
							-- affects this icon and see if there's anyone who can cast it
							local spells = module[self.mode][spell2.key].spells
							local canCast = 0
							if (spells) then
								for name,spellInfo in pairs(spells) do
									local canDo = module.spellPotential[name]
									if (canDo) then							-- Quick check first
										-- Now count who can do it
										if (spellInfo.required) then
											-- Spell requires a talent, so check who can cast it
											for unit in module:IterateRoster() do
												if (not UnitIsUnit(unit, caster)) then
													local _, class = UnitClass(unit)
													if (class == spellInfo.class) then
														if (module:HasTalent(unit, spellInfo.required)) then
															if (not (spellInfo.exclusive or spellInfo.totem) or not module:HasExclusiveAura(self.mode, self.unit, spellInfo.exclusive or spellInfo.totem, UnitGUID(unit))) then
																canCast = canCast + 1
															end
														end
													end
												end
											end
										else
											-- No requirements, add classcount
											canCast = canCast + module.classcount[spellInfo.class]
											if (spellInfo.exclusive or spellInfo.totem) then
												-- And remove how many auras of this class there are active
												canCast = canCast - module:ExclusiveAuraCount(self.mode, self.unit, spellInfo.exclusive or spellInfo.totem)
											end
										end
									end
								end
							end

							if (canCast == 0) then
								icon2:Unavailable(true)
							end
						end
					end
				end
			end
		end

		local mappings = module.mappingLookup[name]
		if (mappings) then
			for i,mapping in ipairs(mappings) do
				if (not mapping.requiredTalent or (caster and module:HasTalent(caster, mapping.requiredTalent))) then
--@debug@
					assert(mapping.key ~= name)			-- Would trigger a stack overflow
--@no-debug@
					on = on + self:AddActiveAura(mapping.key, caster, count)
				end
			end
		end

		return on
	end

	-- UpdateBuffs
	function frame:UpdateBuffs()
		self:AllOff()

		local unit, auraType
		unit = self.unit
		auraType = self.mode == "buffs" and "HELPFUL" or "HARMFUL"

		local on = 0
		for i = 1,1000 do
			local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = UnitAura(unit, i, auraType)
			if (name == nil) then
				break
			end

			on = on + self:AddActiveAura(name, caster, count)
		end

		if (self.bar) then
			self.bar:SetMinMaxValues(0, max(on, InCombatLockdown() and self.totalAvailable or self.totalAvailableNonTemporary))
			self.bar:SetValue(on)
			local r, g, b = SmoothColour(on / max(on, InCombatLockdown() and self.totalAvailable or self.totalAvailableNonTemporary))
			self.bar:SetStatusBarColor(r, g, b)
			self.bar.bg:SetVertexColor(r, g, b, 0.3)
		end
	end

	function frame:SetTitle()
		if (self.backdrop) then
			local gotTalents = module:GetTalentCount()
			local raidSize = GetNumRaidMembers()
			if (raidSize == 0) then
				raidSize = GetNumPartyMembers() + 1
			end
			if (gotTalents < raidSize) then
				self.backdrop.title:SetText(Gradient(format(L["Talents %d/%d"], gotTalents, raidSize), 1, 0.1, 0.1, 1, 1, 0.1))
				return
			else
				if (UnitExists(self.unit) and not UnitIsUnit("player", self.unit)) then
					if (UnitInParty(self.unit) or UnitInRaid(self.unit) or not UnitIsFriend("player", self.unit)) then
						self.backdrop.title:SetText(UnitName(self.unit))
						if (UnitIsPlayer(self.unit)) then
							local _, class = UnitClass(self.unit)
							local c = class and RAID_CLASS_COLORS[class]
							if (c) then
								self.backdrop.title:SetTextColor(c.r, c.g, c.b)
							else
								self.backdrop.title:SetTextColor(module:ReactionColour(self.unit))
							end
						else
							self.backdrop.title:SetTextColor(module:ReactionColour(self.unit))
						end
						return
					end
				end
			end

			self.backdrop.title:SetText(module.label)
			self.backdrop.title:SetTextColor(1, 1, 1)
		end
	end

	frame:SetMode(db.mode or "debuffs")
	self:SetFrameSize()
	frame:Hide()

	self.InitFrames = nil
	self.MakeBaseFrame = nil
end

-- GetAuraForName
-- Like UnitAura, but also checks mappings
-- *** RECURSIVE ***
function module:GetAuraForName(unit, name, unitAuraFunc)
	local auraName, auraRank, auraTex, auraCount, auraDebuffType, auraDuration, auraEndTime, auraCaster, auraIsStealable = unitAuraFunc(unit, name)
	if (auraName) then
		auraRank = tonumber(auraRank:match("(%d+)"))
	else
		local mapping = self.mapping[name]
		if (mapping) then
			-- Check for mappings
			for mappingName in pairs(mapping.list) do
				local valid
				auraName, auraRank, auraTex, auraCount, auraDebuffType, auraDuration, auraEndTime, auraCaster, auraIsStealable = self:GetAuraForName(unit, mappingName, unitAuraFunc)
				if (auraName and (auraName ~= mappingName or (not mapping.requiredTalent or (auraCaster and self:HasTalent(auraCaster, mapping.requiredTalent))))) then
					valid = true
				else
					auraName, auraCaster = nil, nil
				end
				if (valid) then
					if (type(auraRank) == "string") then
						auraRank = tonumber(auraRank:match("(%d+)"))
					end
					break
				end
			end
		end
	end
	return auraName, auraRank, auraTex, auraCount, auraDebuffType, auraDuration, auraEndTime, auraCaster, auraIsStealable
end

-- GetActiveAurasForKey
function module:GetActiveAurasForKey(unit, mode, key)
	info = module[mode][key]					-- DO NOT LOCALIZE, it's up there with the sortList func
	if (not info) then
		return
	end
	local unitAuraFunc = mode == "buffs" and UnitBuff or UnitDebuff

	local list = new()
	for name,spellInfo in pairs(info.spells) do
		tinsert(list, name)
	end
	sort(list, sortList)

	local active, missing
	for i = 1,#list do
		local name = list[i]
		local spellInfo = info.spells[name]

		local auraName, auraRank, auraTex, auraCount, auraDebuffType, auraDuration, auraEndTime, auraCaster, auraIsStealable = self:GetAuraForName(unit, name, unitAuraFunc)
		if (not auraName and spellInfo.alternate) then
			auraName, auraRank, auraTex, auraCount, auraDebuffType, auraDuration, auraEndTime, auraCaster, auraIsStealable = unitAuraFunc(unit, spellInfo.alternate)
			if (auraName) then
				auraRank = tonumber(auraRank:match("(%d+)"))
			end
		end

		if (auraName) then
			if (not active) then
				active = new()
			end
			local n = new()
			active[name] = n
			n.name = auraName
			n.rank = auraRank
			n.tex = auraTex
			n.count = auraCount
			n.type = auraDebuffType
			n.duration = auraDuration
			n.endTime = auraEndTime
			n.caster = auraCaster
			n.isStealable = auraIsStealable
		else
			if (not missing) then
				missing = new()
			end
			missing[name] = true
		end
	end

	return active, missing
end

-- ExclusiveAuraCount
function module:ExclusiveAuraCount(mode, unit, exclusive)
	assert(exclusive)

	local total = 0
	local unitAuraFunc = mode == "buffs" and UnitBuff or UnitDebuff
	for i = 1,1000 do
		local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = unitAuraFunc(unit, i)
		if (not name) then
			break
		end

		local info = self.lookup[mode][name]
		if (info) then
			if (info.exclusive == exclusive or info.totem == exclusive) then
				total = total + 1
			end
		else
			info = self.otherBuffs[name]
			if (info and (info.exclusive == exclusive or info.totem == exclusive)) then
				total = total + 1
			end
		end
	end
	return total
end

-- HasExclusiveAura
function module:HasExclusiveAura(mode, unit, exclusive, guid)
	assert(exclusive)

	local unitAuraFunc = mode == "buffs" and UnitBuff or UnitDebuff

	for i = 1,1000 do
		local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = unitAuraFunc(unit, i)
		if (not name) then
			break
		end
		if (caster) then
			local casterGUID = UnitGUID(caster)
			if (casterGUID == guid) then
				local info = self.lookup[mode][name]
				if (info) then
					if (info.exclusive == exclusive or info.totem == exclusive) then
						return info.id
					end
				else
					info = self.otherBuffs[name]
					if (info and (info.exclusive == exclusive or info.totem == exclusive)) then
						return info.id
					end
				end
			end
		end
	end	
end

-- ReportMissingIcon
function module:ReportMissingIcon(testunit, mode, key)
	if (select(2, IsInInstance()) == "pvp") then
		return
	end
	info = module[mode][key]							-- DO NOT LOCALIZE, it's up there with the sortList func
	if (not info or not info.reportable) then
		return
	end

	if (mode == "debuffs") then
		if (UnitIsDead(testunit) or not UnitCanAttack("player", testunit)) then
			return
		end
	end

	local active, missing = module:GetActiveAurasForKey(testunit, mode, key)
	if (active) then
		-- Some active spell for this key, so nothing to ask anyone
		deepDel(active)
		del(missing)
		return
	end

	local list = new()
	for name,spellInfo in pairs(info.spells) do
		tinsert(list, name)
	end
	sort(list, sortList)

	local testtype = mode == "buffs" and L["buff"] or L["debuff"]
	local bestUnit, bestSpell, bestAmount, couldCast
	for i = 1,#list do
		local name = list[i]
		local spellInfo = info.spells[name]

		local canDo = module.spellPotential[name]
		if (not canDo) then
			if (petClasses[spellInfo.class]) then
				canDo = module.petPotential and module.petPotential[name]
			end
		end
		local div = canDo and 1 or 2

		if (canDo and (not active or not active[name])) then
			assert(missing and missing[name])

			for unit in self:IterateRoster() do
				local _, class = UnitClass(unit)
				if (class == spellInfo.class) then
					local canCast
					if (spellInfo.requiredTalent) then
						canCast = self:HasTalent(unit, spellInfo.requiredTalent)
					else
						canCast = true
					end

					if (canCast) then
						if (spellInfo.exclusive or spellInfo.totem) then
							local id = self:HasExclusiveAura(mode, testunit, spellInfo.exclusive or spellInfo.totem, UnitGUID(unit))
							if (id) then
								if (not couldCast) then
									couldCast = new()
								end

								couldCast[unit] = id
								canCast = nil
							end
						end

						if (canCast) then
							if (spellInfo.improved) then
								if (self:HasTalent(unit, spellInfo.improved)) then
									local amount = self:GetAmount(spellInfo, true, true)
									if (not bestAmount or amount > bestAmount) then
										bestUnit = unit
										bestSpell = spellInfo
										bestAmount = amount
									end
								end
							else
								local amount = self:GetAmount(spellInfo, true)
								if (not bestAmount or amount > bestAmount) then
									bestUnit = unit
									bestSpell = spellInfo
									bestAmount = amount
								end
							end
						end
					end					
				end
			end
		end
	end

	if (bestUnit) then
		local spellLink = self:LinkSpell(bestSpell.id)
		if (UnitIsUnit(bestUnit, "player")) then
			if (UnitIsUnit(testunit, "player")) then
				self:Print(format(L["You are missing your own %s %s"], spellLink, testtype))
			else
				self:Print(format(L["%s is missing your own %s %s"], self:ColourPlayer(testunit), spellLink, testtype))
			end
		else
			self:Print(L["Notifing %s regarding %s"], self:ColourPlayer(bestUnit), spellLink)
			if (UnitIsUnit(testunit, "player")) then
				SendChatMessage("Utopia> "..format(L["I am missing your %s %s"], spellLink, testtype), "WHISPER", nil, UnitFullName(bestUnit))
			else
				SendChatMessage("Utopia> "..format(L["%s is missing your %s %s"], UnitFullName(testunit), spellLink, testtype), "WHISPER", nil, UnitFullName(bestUnit))
			end
		end
	else
		self:Print(format(L["There is noone to cast |cFFFFFF80%s|r"], key))
		if (couldCast) then
			for unit,spellId in pairs(couldCast) do
				local spell
				if (type(spellId) == "string") then
					spell = spellId
				else
					spell = self:LinkSpell(spellId)
				end
				self:Print(format("%s could cast this but currently has %s active", self:ColourPlayer(unit), spell))
			end
		end
	end

	del(couldCast)
	deepDel(active)
	del(missing)
end

-- LinkSpell
function module:LinkSpell(id)
	return format("|cff71d5ff|Hspell:%d|h[%s]|h|r", id, (GetSpellInfo(id)))
end

do
	local function chatFilterInform(self, event, ...)
		local msg = ...
		if (strfind(msg, "^Utopia>")) then
			return true, ...
		end
		return false, ...
	end

	-- HookChat
	function module:HookChat()
		ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", chatFilterInform)
	end

	-- z:UnhookChat
	function module:UnhookChat()
		ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER_INFORM", chatFilterInform)
	end
end

local CountTree
do
	local count
	local function CountTreeBits(c)
		count = count + tonumber(c)
	end
	function CountTree(branch)
		count = 0
		branch:gsub(".", CountTreeBits)
		return count
	end
end

-- CheckRoleChange
function module:CheckRoleChange()
	local role = self.role
	self:MyRole(true)
	if (self.role ~= role) then
		if (self.frame) then
			self.frame:SetMode()
		end
	end
end

-- MyRole
function module:MyRole(reset)
	if (self.role and not reset) then
		return self.role
	end

	self.role = nil
	if (playerClass == "ROGUE" or playerClass == "HUNTER") then
		self.role = "melee"
	elseif (playerClass == "MAGE" or playerClass == "WARLOCK") then
		self.role = "caster"
	elseif (playerClass == "DEATHKNIGHT") then
		self.role = UnitAura("player", GetSpellInfo(48263)) and "tank" or "melee"				-- Frost Presence
	else
		local t = self.talents[playerName]
		if (t) then
			local t1, t2, t3 = CountTree(t[1]), CountTree(t[2]), CountTree(t[3])
			local heavy = (t1 > t2 and t1 > t3 and 1) or (t2 > t1 and t2 > t3 and 2) or (t3 > t1 and t3 > t2 and 3) or 0

			if (playerClass == "PALADIN") then
				self.role = heavy == 1 and "healer" or heavy == 2 and "tank" or heavy == 3 and "melee"

			elseif (playerClass == "PRIEST") then
				self.role = heavy == 3 and "caster" or "healer"

			elseif (playerClass == "WARRIOR") then
				self.role = heavy == 3 and "tank" or "melee"

			elseif (playerClass == "DRUID") then
				if (heavy == 2) then
					if (UnitAura("player", GetSpellInfo(5487)) or UnitAura("player", GetSpellInfo(9634))) then
						-- Bear Forms = tank
						self.role = "tank"
					else
						self.role = "melee"
					end
				else
					self.role = heavy == 1 and "caster" or heavy == 3 and "healer"
				end

			elseif (playerClass == "SHAMAN") then
				self.role = heavy == 1 and "caster" or heavy == 2 and "melee" or heavy == 3 and "healer"
			end
		end
	end
	return self.role
end

-- ReactionColour
function module:ReactionColour(unit)
	if (not UnitPlayerControlled(unit)) then
		if (UnitIsDeadOrGhost(unit)) then
			return 0.4, 0.4, 0.4
		elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) then
			return 0.6, 0.6, 0.6
		else
			local reaction = UnitReaction(unit, "player")
			if (reaction) then
				if (reaction >= 5) then
					return 0, 1, 0
				elseif (reaction <= 2) then
					return 1, 0, 0
				elseif (reaction == 3) then
					return 1, 0.5, 0
				else
					return 1, 1, 0
				end
			end
		end
	end
	return 0, 0.5, 1
end

-- SetFrameSize
function module:SetFrameSize()
	local frame = self.frame
	if (frame) then
		local db = db
		local cols = db.columns
		local rows = ceil(frame.shown / cols)
		local size = db.size

		if (db.orientation == "H") then
			frame:SetWidth(size * cols + (db.hspacing * (cols - 1)))
			frame:SetHeight(size * rows + (db.vspacing * (rows - 1)) + (db.bar and db.barwidth or 0))
		else
			frame:SetWidth(size * rows + (db.hspacing * (rows - 1)) + (db.bar and db.barwidth or 0))
			frame:SetHeight(size * cols + (db.vspacing * (cols - 1)))
		end
	end
end

-- SetBarOrientation
function module:SetBarOrientation(info, n)
	if (n == "B" or n == "T") then
		db.side = n
	end

	local frame = self.frame
	if (frame) then
		self:SetFrameSize()

		if (frame.bar) then
			frame.bar:ClearAllPoints()
			if (db.orientation == "H") then
				if (db.side == "B") then
					frame.bar:SetPoint("BOTTOMLEFT")
					frame.bar:SetPoint("TOPRIGHT", frame, "BOTTOMRIGHT", 0, db.barwidth)
				else
					frame.bar:SetPoint("TOPLEFT")
					frame.bar:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 0, -db.barwidth)
				end
			else
				if (db.side == "B") then
					frame.bar:SetPoint("TOPRIGHT")
					frame.bar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", -db.barwidth, 0)
				else
					frame.bar:SetPoint("TOPLEFT")
					frame.bar:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", db.barwidth, 0)
				end
			end
		end
	end
end

-- SetOrientation
function module:SetOrientation(info, newValue)
	local n
	if (newValue == "V" or newValue == "H") then
		db.orientation = newValue
	elseif (newValue == "T" or newValue == "B") then
		db.side = newValue
	end
	local n = db.orientation

	local frame = self.frame
	if (frame) then
		frame.shown = 0
		local prev, prevRow
		for i,icon in ipairs(frame.icons) do
			if (icon:IsShown()) then
				frame.shown = frame.shown + 1
				icon:ClearAllPoints()
				if (prev) then
					if ((frame.shown - 1) % db.columns == 0) then
						if (n == "H") then
							icon:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT", 0, -db.vspacing)
						else
							icon:SetPoint("TOPLEFT", prevRow, "TOPRIGHT", db.hspacing, 0)
						end
						prevRow = icon
					else
						if (n == "H") then
							icon:SetPoint("TOPLEFT", prev, "TOPRIGHT", db.hspacing, 0)
						else
							icon:SetPoint("TOPLEFT", prev, "BOTTOMLEFT", 0, -db.vspacing)
						end
					end
				else
					if (db.side == "T") then
						if (n == "H") then
							icon:SetPoint("TOPLEFT", 0, -db.barwidth)
						else
							icon:SetPoint("TOPLEFT", db.barwidth, 0)
						end
					else
						icon:SetPoint("TOPLEFT")
					end
					prevRow = icon
				end
				prev = icon
			end
		end
		self:SetBarOrientation()
		self:UpdateBackdrop()
	end
end

-- UpdateBackdrop
local bgFrame = {insets = {}}
function module:UpdateBackdrop()
	if (db.Background.Enable) then
		bgFrame.bgFile = SM:Fetch("background", db.Background.Texture)
		bgFrame.edgeFile = SM:Fetch("border", db.Background.BorderTexture)
		bgFrame.tile = db.Background.Tile
		bgFrame.tileSize = db.Background.TileSize
		bgFrame.edgeSize = db.Background.EdgeSize
		local inset = floor(db.Background.EdgeSize / 4)
		bgFrame.insets.left = inset
		bgFrame.insets.right = inset
		bgFrame.insets.top = inset
		bgFrame.insets.bottom = inset
	
		local bd = self.frame.backdrop
		if (not bd) then
			-- Backdrop is going to be larger than the frame proper
			local bd = CreateFrame("Frame", nil, self.frame)
			self.frame.backdrop = bd
			bd:SetFrameLevel(self.frame:GetFrameLevel() - 1)
			--self.frame:SetFrameLevel(bd:GetFrameLevel() + 2)
			bd:SetScript("OnDragStart", dragStart)
			bd:SetScript("OnDragStop", dragStop)
			bd:SetScript("OnEnter", function() module:OnEnter() end)
			bd:SetScript("OnLeave", function() module:OnLeave() end)
			bd:RegisterForDrag("LeftButton")
			if (not db.locked) then
				bd:EnableMouse(true)
			end

			bd.title = bd:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			bd.title:SetText(self.label)
			bd.title:SetPoint("BOTTOMLEFT", self.frame, "TOPLEFT", -1, 1)
			bd.title:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT", 1, 11)
			bd.title:SetJustifyH("LEFT")
		end

		local esd2 = db.Background.EdgeSize / 2
		self.frame.backdrop:ClearAllPoints()
		self.frame.backdrop:SetPoint("TOPLEFT", -esd2, esd2)
		self.frame.backdrop:SetPoint("BOTTOMRIGHT", esd2, -esd2)
	
		self.frame.backdrop:SetBackdrop(bgFrame)
	
		local c = db.Background.Color
		self.frame.backdrop:SetBackdropColor(c.r, c.g, c.b, c.a)
	
		c = db.Background.BorderColor
		self.frame.backdrop:SetBackdropBorderColor(c.r, c.g, c.b, c.a)

		self.frame.backdrop:Show()
	else
		if (self.frame.backdrop) then
			self.frame.backdrop:Hide()
		end
	end
end

-- ToggleActive
function module:ToggleActive()
	db.enable = not db.enable
	self:ShowWhen()
end

-- SetIconSize
function module:SetIconSize(info, n)
	db.size = n
	
	local frame = self.frame
	if (self.frame) then
		for i,icon in pairs(self.frame.icons) do
			icon:SetWidth(n)
			icon:SetHeight(n)
		end

		self:SetFrameSize()
		self:SavePosition(module.frame)
	end
end

-- SetScale
function module:SetScale()
	if (self.frame) then
		self.frame:SetScale(db.scale)
	end
end

-- ShowWhen
function module:ShowWhen()
	if (db.enable) then
		if (not self.frame) then
			self:InitFrames()
		end

		local show = db.show.always
		if (not show) then
			local pvpInstance = select(2, IsInInstance()) == "pvp"
			if (self.pvp or not pvpInstance) then
				if (GetNumRaidMembers() > 0) then
					show = db.show.raid
				elseif (GetNumPartyMembers() > 0) then
					show = db.show.party
				else
					show = db.show.solo
				end
			end
		end
		if (show) then
			if (UnitCanAttack("player", "target") and not UnitIsDead("target")) then
				if (db.show.enemy) then
					show = true
				elseif (db.show.boss) then
					if (UnitClassification(self.frame.unit) == "worldboss") then
						show = true
					else
						show = false
					end
				end
			else
				show = not db.show.enemy and not db.show.boss
			end
		end

		if (show) then
			if (db.automode) then
				if (self.frame.mode == "buffs") then
					if (UnitCanAttack("player", "target") and not UnitIsDead("target")) then
						self.frame:SetMode("debuffs")
					end
				else
					if (not UnitExists("target") or UnitIsUnit("player", "target") or (UnitIsFriend("player", "target") and db.allPlayers) or UnitIsDead("target")) then
						self.frame:SetMode("buffs")
					end
				end
			end

			self.frame:Show()
			self.frame:SetTitle()
		else
			self.frame:Hide()
		end

	elseif (self.frame) then
		self.frame:Hide()
	end
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	self:CheckForMissingTalents()
	self:ShowWhen()
end

-- UNIT_LEVEL
function module:UNIT_LEVEL(unit)
	if (UnitInParty(unit) or UnitInRaid(unit)) then
		self:ScanRoster()
	end
end

-- PARTY_MEMBER_ENABLE
function module:PARTY_MEMBER_ENABLE()
	self:CheckForMissingTalents()
end

-- PARTY_MEMBERS_CHANGED
function module:PARTY_MEMBERS_CHANGED()
	self.rosterUpdate = true
end

-- OnRaidRosterUpdate
function module:OnRaidRosterUpdate()
	self.distribution = nil
	if (self.pvp or select(2, IsInInstance()) ~= "pvp") then
		if (GetNumRaidMembers() > 0) then
			self.distribution = "RAID"
		elseif (GetNumPartyMembers() > 0) then
			self.distribution = "PARTY"
		else
			self.sentHello = nil
		end
	end

	if (self.distribution) then
		if (not self.sentHello or (self.sentHello < GetTime() - 5 and not self.users)) then
			self.sentHello = GetTime()
			self:SendCommMessage(commPrefix, "HELLO "..self.version, self.distribution)
		end
	end

	self:ScanRoster()
	if (self.distribution or db.show.always or db.show.solo) then
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
	else
		self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	end

	self:PLAYER_TARGET_CHANGED()
end

do
	local function iter(t)
		local key = t.id
		local ret
		if (t.mode == "raid") then
			if (key > t.r) then
				del(t)
				return nil
			end
			ret = "raid"..key
		else
			if (key > t.p) then
				del(t)
				return nil
			end
			ret = key == 0 and "player" or "party"..key
		end
		t.id = key + 1
		return ret
	end

	-- IterateRoster
	function module:IterateRoster()
		local t = new()
		if (GetNumRaidMembers() > 0) then
			t.mode = "raid"
			t.id = 1
			t.r = GetNumRaidMembers()
		else
			t.mode = "party"
			t.id = 0
			t.p = GetNumPartyMembers()
		end
		return iter, t
	end
end

-- GetTalentCount
function module:GetTalentCount()
	local count = 0
	for name in pairs(self.talents) do
		count = count + 1
	end
	return count
end

-- GetTalentMissingPeople
function module:GetTalentMissingPeople()
	local list = new()
	for unit in self:IterateRoster() do
		local name = UnitFullName(unit)
		if (name and name ~= UNKNOWN) then
			if (not self.talents[name]) then
				tinsert(list, name)
			end
		end
	end

	local missing
	if (#list > 0) then
		sort(list)
		for i,name in ipairs(list) do
			list[i] = self:ColourPlayer(name)
		end
		missing = table.concat(list, ", ")
	end
	del(list)

	return missing
end

-- ScanRoster
function module:ScanRoster()
	if (not self.talents) then
		self.talents = {}
	end
	self.potential.buffs = del(self.potential.buffs)
	self.potential.debuffs = del(self.potential.debuffs)
	self.potential.buffs = new()
	self.potential.debuffs = new()
	del(self.spellPotential)
	self.spellPotential = new()
	self.petPotential = deepDel(self.petPotential)
	self.pets = del(self.pets)
	self.petlookup = del(self.petlookup)
	self.temporary = deepDel(self.temporary)

	self.temporary = new()
	self.temporary.buffs = new()
	self.temporary.debuffs = new()

	local oldTalents = new()
	for name,talents in pairs(self.talents) do
		oldTalents[name] = true
	end

	self.classcount = setmetatable({}, {__index = function() return 0 end})
	for unit in self:IterateRoster() do
		if (oldTalents) then
			local name = UnitFullName(unit)
			oldTalents[name] = nil
		end

		local _, class = UnitClass(unit)
		if (class) then
			self.classcount[class] = self.classcount[class] + 1

			local talents = self:GetTalents(unit)

			self:AddPotential(unit, talents)

			if (petClasses[class]) then
				self:UpdateUnitPet(unit, true)
			end
		end
	end

	if (oldTalents) then
		for name in pairs(oldTalents) do
			self.talents[name] = del(self.talents[name])
			if (self.users) then
				self.users[name] = nil
			end
			if (throttle) then
				throttle[name] = nil
			end
		end
		del(oldTalents)
		if (self.users and not next(self.users)) then
			self.users = del(self.users)
		end
		if (throttle and not next(throttle)) then
			throttle = del(throttle)
		end
	end

	self:MakePetPotentialKeys()
	if (self.frame) then
		self.frame:UpdateBuffs()
	end
end

-- UNIT_PET
function module:UNIT_PET(event, ownerunit)
	self:UpdateUnitPet(ownerunit)
end

-- PurgePetsForOwner
function module:PurgePetsForOwner(owner)
	local ret
	if (self.petlookup) then
		for petguid,ownerName in pairs(self.petlookup) do
			if (ownerName == owner) then
				self.petlookup[petguid] = nil
				ret = true
				break
			end
		end
		if (not next(self.petlookup)) then
			self.petlookup = del(self.petlookup)
--@debug@
		else
			for petguid,ownerName in pairs(self.petlookup) do
				if (ownerName == owner) then
					error("Unexpected pet registered to %s", self:ColourPlayer(owner))
				end
			end
--@no-debug@
		end
	end
	return ret
end

-- UpdateUnitPet
function module:UpdateUnitPet(unit, noPetUpdate)
	local owner = UnitFullName(unit)
	local petid
	if (unit == "player") then
		petid = "pet"
	else
		petid = unit:gsub("^(%a+)(%d+)$", "%1pet%2")
	end
	local petguid = UnitGUID(petid)

	-- Maintain the pet->owner lookup
	local okpet = UnitExists(petid)
	if (okpet and self.petlookup and self.petlookup[petguid] == owner) then
		-- No change. ie: Superfluous UNIT_PET event. They happen often
		return
	end
	if (not self:PurgePetsForOwner(owner) and not okpet) then
		-- Also no change, pet was absent and still is
		return
	end
	if (okpet) then
		if (not self.petlookup) then
			self.petlookup = new()
		end
		self.petlookup[petguid] = owner
	end

	-- Now check type of pet and see what has to be made available or removed
	local _, class = UnitClass(unit)
	if (petClasses[class]) then
		local cType = UnitCreatureFamily(petid)				-- UnitExists(petid) and 
		local remove
		if (cType) then
			if (not self.pets) then
				self.pets = new()
			end
			if (self.pets[owner] == cType) then
				return				-- No change
			end

			-- 'owner' now has a new pet of type 'cType'
			self.pets[owner] = cType
		else
			if (self.pets) then
				if (not self.pets[owner]) then
					return				-- No change
				end
				cType = self.pets[owner]				-- Set this so we know what to remove in next section
				remove = true

				-- 'owner' had a pet of 'cType', but lost it
				self.pets[owner] = nil
				if (not next(self.pets)) then
					self.pets = del(self.pets)
				end
			end
		end

		if (cType) then
			local change
			local c = self.classes[class]
			if (c) then
				for buffsDebuffs,buffList in pairs(c) do
					for key, name in pairs(buffList) do
						local info = self.lookup.debuffs[name] or self.lookup.buffs[name]		-- Assumption

						if (info and info.pet == "any" or info.pet == cType) then
							local list = self.petPotential and self.petPotential[name]

							if (remove) then
								if (list and list[owner]) then
									-- Removing pet triggered aura (name) from player (owner)
									list[owner] = nil
									if (not next(list)) then
										self.petPotential[name] = del(list)
										change = true
									end
								end
							else
								if (not self.petPotential) then
									self.petPotential = new()
								end
								if (not list) then
									list = new()
									self.petPotential[name] = list
								end

								-- Adding pet triggered aura (name) to player (owner)
								if (not list[owner]) then
									change = true
									list[owner] = true
								end
							end
						end
					end
				end
			end
		end
		if (change) then
			if (not noPetUpdate) then
				self:MakePetPotentialKeys()
				if (self.frame) then
					self.frame:UpdateBuffs()
				end
			end
		end
	end
end

-- MakePetPotentialKeys
function module:MakePetPotentialKeys()
	self.petPotentialKeys = del(self.petPotentialKeys)

	if (self.petPotential) then
		for spellName,list in pairs(self.petPotential) do
			local spell = self.lookup.debuffs[spellName] or self.lookup.buffs[spellName]
			if (not self.petPotentialKeys) then
			 	self.petPotentialKeys = new()
			end
			self.petPotentialKeys[spell.key] = true
		end
	end
end

-- AddPotentialSpells
function module:AddPotentialSpells(unit, talents, class, Type)
	local unitLevel = UnitLevel(unit)
	for category, spells in pairs(self[Type]) do
		for spellName, info in pairs(spells.spells) do
			if (info.class == class) then
				if (not info.minLevel or unitLevel >= info.minLevel) then
					if (not info.rankLevels or unitLevel >= info.rankLevels[1]) then
						local needsPet = info.pet	-- and petClasses[class]

						if (not info.requiredTalent and not needsPet) then
							local level = 0
							if (info.improved) then
								level = self:HasTalent(unit, info.improved) or 0
							end

							self.potential[Type][info.key] = max(self.potential[Type][info.key] or 0, level)
							self.spellPotential[spellName] = max(self.spellPotential[spellName] or 0, level)

							if (info.temporary or (info.totem and db.totemsTemporary)) then
								self.temporary[Type][spellName] = true
							end
						else
							if (talents) then
								local level = self:HasTalent(unit, info.requiredTalent)
								if (level and level > 0) then
									if (info.improved) then
										level = self:HasTalent(unit, info.improved) or level
									end

									if (info.maxTalentPoints and level > info.maxTalentPoints) then
										self.talents[UnitFullName(unit)] = nil			-- Assume bad talent scan
									else
										if (not needsPet) then
											self.potential[Type][info.key] = max(self.potential[Type][info.key] or 0, level)
											self.spellPotential[spellName] = max(self.spellPotential[spellName] or 0, level)
										end

										if (info.temporary or (info.totem and db.totemsTemporary)) then
											self.temporary[Type][spellName] = true
										end
									end
								end
							end
						end
					end
				end
			end
		end
	end
end

-- AddPotential
function module:AddPotential(unit, talents)
	local _, class = UnitClass(unit)
	if (class) then
		self:AddPotentialSpells(unit, talents, class, "debuffs")
		self:AddPotentialSpells(unit, talents, class, "buffs")
	end
end

-- HasTalent
function module:HasTalent(unit, required)
	if (unit and self.classTalentData) then
		local talents = self.talents[UnitFullName(unit)]
		if (talents) then
			local _, class = UnitClass(unit)
			if (class) then
				local data = self.classTalentData[class]
				if (data) then
					local info = data.list and data.list[required]
					if (info) then
						local amount = talents[info.treeIndex]:byte(info.index) - 48
						if (amount) then
							if (amount > 0) then
								return amount
							end
--@debug@
else
d("Problem: Utopia:HasTalent(%s, %s) - %s", tostring(unit), tostring(required), tostring(talents[tree]))
--@no-debug@
						end
					end
				end
			end
		end
	end
end

-- OnReceiveTalents
function module:OnReceiveTalents(fullname, talents)
	if (talents and next(talents) and type(talents[1]) == "string") then
		-- We validate the talents we've received to be of specific lengths.
		-- There are cases when the Lib gets it wrong and gives us a notify for wrong people.
		local err
		local _, class = UnitClass(fullname)
		if (class) then
			local lens = self.expectedTalentLengths[class]
			if (lens) then
				for i = 1,#talents do
					if (lens[i] ~= #talents[i]) then
--@debug
d("Bad talent read for %s. Tree |cFFFFFF80%d|r expected length |cFF80FF80%d|r, got |cFFFF8080%d|r. Ignored", self:ColourPlayer(fullname), i, lens[i], #talents[i])
--@no-debug@
						err = true
					end
				end
			end
		end

		if (not err) then
			if (db.storeGuildTalents) then
				if (GetNumRaidMembers() > 0) then
					if (GetGuildInfo(fullname) == GetGuildInfo("player") and not UnitIsUnit(fullname, "player")) then
						if (not db.guildTalents) then
							db.guildTalents = {}
						end
						db.guildTalents[fullname] = copy(talents)
					end
				end
			end

			if (self.frame and self.frame:IsShown()) then
				self.frame:SetTitle()
			end

			if (self.missingTalents) then
				self.missingTalents[fullname] = nil
			end

			self:AddPotential(fullname, talents)
			if (self.frame) then
				self.frame:UpdateBuffs()

				if (fullname == playerName) then
					self:CheckRoleChange()
				end
			end
		end
	end
end

-- GetTalents
do
	local RockComm = LibStub("LibRockComm-1.0", true)
	if (RockComm) then
		local function talentReceptor(sender, data)
			if (type(data) == "table" and #data == 3 and type(data[1]) == "table") then
				--self:Print("GOT TALENTS from RockComm for "..name)
				local n = new()
				
				for i = 1,3 do
					local str = ""
					for j = 1,#data[i] do
						str = str .. tostring(data[i][j])
					end
					tinsert(n, str)
				end
				module.talents[sender] = n
				module:OnReceiveTalents(sender, n)
			end
		end
		RockComm:AddTalentReceptor(talentReceptor)
	end

	-- GetTalents
	function module:GetTalents(unit)
		local name = UnitFullName(unit)
		if (name == UNKNOWN) then
			return
		end
		local talents = self.talents[name]
		local refetch

		if (not talents) then
			-- Check for stored guildy talents
			talents = db.guildTalents and db.guildTalents[name]
			if (talents) then
				-- Always refetch these
				self.talents[name] = copy(talents)
				refetch = true
			end
		else
			-- If someone is out of sight, we won't catch their talent swap spell cast, so we'll invalidate them here and recheck talents
			if (talents and ((not UnitIsVisible(unit) and UnitIsConnected(unit)) or (self.outOfSight and self.outOfSight[name]))) then
				if (not self.users or type(self.users[name]) ~= "number") then
					refetch = true
				end
			end
		end

		if ((not talents or refetch) and not self.talentRecursion) then
			if (not self.missingTalents) then
				self.missingTalents = new()
			end
			self.missingTalents[name] = true

			if (UnitIsUnit("player", unit)) then
				talents = self:TalentQuery_Ready(nil, UnitFullName(unit))

			elseif ((UnitInRaid(unit) or UnitInParty(unit)) and UnitIsConnected(unit)) then
				TalentQuery:Query(unit)

				if (not UnitIsVisible(unit) or not CanInspect(unit)) then
					if (self.users and self.users[name]) then
						if (not self.commQueried or not self.commQueried[name]) then
							if (not self.commQueried) then
								self.commQueried = new()
							end
							self.commQueried[name] = true
							self:SendCommMessage(commPrefix, "REQUESTTALENTS", "WHISPER", name)
						end

					elseif (RockComm) then
						if (not self.rockCommQueried or not self.rockCommQueried[name]) then
							if (not self.rockCommQueried) then
								self.rockCommQueried = new()
							end
							self.rockCommQueried[name] = true
							RockComm:QueryTalents("WHISPER", name)
						end
					end
				end
			end
		end

		return talents
	end
end

-- CheckForMissingTalents
function module:CheckForMissingTalents()
	if (self.missingTalents) then
		if (self.loggedIn and UnitInParty("player")) then
			for name in pairs(self.missingTalents) do
				-- We can use IsVisible, becuase we've already tried once to request
				-- talents via comms and failed, So we need them close to inspect
				if (UnitInRaid(name) or UnitInParty(name)) then
					if (CanInspect(name)) then
						--self:Print("CheckForMissingTalents - "..name)
						self:GetTalents(name)
					end
				else
					if (name ~= "player" and name ~= playerName) then
						--self:Print("CheckForMissingTalents - removing "..name.." (not in party/raid)")
						self.missingTalents[name] = nil
					end
				end
			end
		end

		if (not next(self.missingTalents)) then
			self.missingTalents = del(self.missingTalents)
		end
	end
end

-- GetClassTalentData
function module:GetClassTalentData(unit)
	local _, class = UnitClass(unit)

	if (class) then
		local data = self.classTalentData and self.classTalentData[class]
		if (not data) then
			local isnotplayer = not UnitIsUnit("player", unit)
			if (GetNumTalentTabs(isnotplayer) > 0) then
				data = new()

				for tab = 1, GetNumTalentTabs(isnotplayer) do
					local tree = new()
					local pointssptent
					tree.name, tree.icon, pointssptent = GetTalentTabInfo(tab, isnotplayer)
					tinsert(data, tree)

					tree.list = new()
					for i = 1,GetNumTalents(tab, isnotplayer) do
						local name, icon, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(tab, i, isnotplayer)
						if (name) then
							local entry = new()
							entry.name = name
							entry.icon = icon
							entry.tier = tier
							entry.column = column
							entry.maxRank = maxRank
							entry.index = i
							entry.treeIndex = tab
							tinsert(tree.list, entry)
							if (not data.list) then
								data.list = new()
							end
							data.list[name] = entry
						end
					end
				end

				if (next(data)) then
					if (not self.classTalentData) then
						self.classTalentData = new()
					end
					self.classTalentData[class] = data
				else
					deepDel(data)
				end
			end
		end
	end
end

-- TalentQuery_Ready
function module:TalentQuery_Ready(e, name, realm)
--self:Print("GOT TALENTS via TalentQuery_Ready for "..name)
	local namerealm = realm and realm ~= "" and name .. "-" .. realm or name

	local isnotplayer = (namerealm ~= playerName)

	self:GetClassTalentData(namerealm)

	local group = GetActiveTalentGroup(isnotplayer)
	if (GetNumTalentTabs(isnotplayer) and GetNumTalentTabs(isnotplayer) >= 3 and GetNumTalents(1, isnotplayer) > 0) then
		local n = new()
		for tab = 1, GetNumTalentTabs(isnotplayer) do
			local t = new()
			for i = 1,GetNumTalents(tab, isnotplayer, nil, group) do
				local name, icon, tier, column, currentRank, maxRank, isExceptional, meetsPrereq = GetTalentInfo(tab, i, isnotplayer, nil, group)
				tinsert(t, currentRank)
			end
			local str = table.concat(t)
			del(t)
			tinsert(n, str)
		end

		self.talents[namerealm] = n
		self:OnReceiveTalents(namerealm, n)

		return n
	end
end

-- TalentQuery_Ready_Outsider
function module:TalentQuery_Ready_Outsider(e, name, realm)
	local namerealm = realm and realm ~= "" and name .. "-" .. realm or name
	if (UnitInRaid(namerealm) or UnitInParty(namerealm)) then
		-- Some other mod has pinged for talents, so we may as well take the opportunity to refresh if we're grouped
		self:TalentQuery_Ready(e, name, realm)
	end
end

-- PLAYER_ENTERING_WORLD
function module:PLAYER_ENTERING_WORLD()
	self.loggedIn = true
	self.auraRefreshAfterLogin = true

	if (not self.initialLogin) then
		self.initialLogin = true
		self:ScheduleTimer("PLAYER_ENTERING_WORLD", 2, self)
	end
	
	self:CheckForMissingTalents()

	if (self.pvp or select(2, IsInInstance()) ~= "pvp") then
		if (self.frame) then
			self.frame:UpdateBuffs()
		end
	end
end

-- PLAYER_LEAVING_WORLD
function module:PLAYER_LEAVING_WORLD()
	self.loggedIn = nil
end

-- SendMyTalents
function module:SendMyTalents(sender)
	local t = self.talents[playerName]
	if (t) then
		local str = format("TALENTS %s,%s,%s", t[1], t[2], t[3])
		if (sender) then
			self:SendCommMessage(commPrefix, str, "WHISPER", sender)
		else
			for name,ver in pairs(self.users) do
				if (type(ver) == "number" and (UnitInRaid(name) or UnitInParty(name))) then
					self:SendCommMessage(commPrefix, str, "WHISPER", name)
				end
			end
		end
	end
end

-- Throttle - Purposely local to here
-- Abuse prevention. Yes, who would abuse addon comms.. Noone would make a macro to crash a mod user would they. right..
local function Throttle(sender, key)
	if (not throttle) then
		throttle = new()
	end
	local s = throttle[sender]
	if (not s) then
		s = new()
		throttle[sender] = s
	end

	local ret = (s[key] or 0) < GetTime() - 5	-- Same message key only allowable once every 15 secs from 1 person
	s[key] = GetTime()
	return ret
end

-- OnCommReceived
function module:OnCommReceived(prefix, msg, channel, sender)
	if (sender == playerName) then
		return
	elseif (not UnitInRaid(sender) and not UnitInParty(sender)) then
		return
	end

	if (prefix == commPrefix) then
		local cmd, str = msg:match("^(%a+) *(.*)$")
		if (not cmd) then
			return
		end

		if (cmd == "TALENTS") then
			if (self.missingTalents) then
				self.missingTalents[sender] = nil
				if (not next(self.missingTalents)) then
					self.missingTalents = del(self.missingTalents)
				end
			end

			local list = new(strsplit(",", str))
			if (next(list)) then
				del(self.talents[sender])
				self.talents[sender] = list
				self:OnReceiveTalents(sender, list)
			else
				del(list)
			end

		elseif (cmd == "REQUESTTALENTS") then
			if (sender ~= playerName) then
				if (Throttle(sender, "REQUESTTALENTS")) then
					self:SendMyTalents(sender)
				end
			end

		elseif (cmd == "HELLO") then
			if (not self.users) then
				self.users = new()
			end
			self.users[sender] = tonumber(str) or str

			if (channel ~= "WHISPER" and sender ~= playerName) then
				if (str ~= "ZOMG" and Throttle(sender, "HELLO")) then
					self:SendCommMessage(commPrefix, "HELLO "..self.version, "WHISPER", sender)
				end
			end
		end
	end
end

-- UserCount
function module:UserCount()
	local count = 0
	if (self.users) then
		for name,ver in pairs(self.users) do
			if (type(ver) == "number") then
				if (UnitInRaid(name) or UnitInParty(name)) then
					count = count + 1
				else
					self.users[name] = nil
				end
			end
		end
		if (not next(self.users)) then
			self.users = del(self.users)
		end
	end
	return count
end

-- RefreshTalents
function module:RefreshTalents(unit)
	local name = UnitFullName(unit or "player")
	self.talents[name] = nil
	if (self.commQueried) then
		self.commQueried[name] = nil
		if (not next(self.commQueried)) then
			self.commQueried = del(self.commQueried)
		end
	end
	if (self.rockCommQueried) then
		self.rockCommQueried[name] = nil
		if (not next(self.rockCommQueried)) then
			self.rockCommQueried = del(self.rockCommQueried)
		end
	end

	self:ScanRoster()
	if (unit == "player") then
		self.talentTimer = nil
	end
end

-- PLAYER_TALENT_UPDATE
function module:PLAYER_TALENT_UPDATE()
	if (self.talentTimer) then
		self:CancelTimer(self.talentTimer)
	end
	self.talentTimer = self:ScheduleTimer("RefreshTalents", 2, "player")
	self:CheckRoleChange()
end

-- GetBloodlustAvailable
function module:GetBloodlustAvailable()
	local satedEnd, freeShaman, bloodLustCooldown, whoCanCast
	local shamans = 0
	local now = GetTime()
	local totalPlayers, playersWithSated = 0, 0

	for unit in self:IterateRoster() do
		if (not UnitIsDeadOrGhost(unit)) then
			totalPlayers = totalPlayers + 1
			local name, rank, tex, count, Type, duration, endTime, caster, isStealable = UnitDebuff(unit, "Sated")
			if (name) then
				playersWithSated = playersWithSated + 1
				if (not satedEnd or endTime < satedEnd) then
					satedEnd = endTime
				end
			end
		end

		if (select(2, UnitClass(unit)) == "SHAMAN" and UnitLevel(unit) >= 70) then
			shamans = shamans + 1
			if (self.usedBloodlust) then
				local guid = UnitGUID(unit)
				local t = self.usedBloodlust[guid]
				if (t) then
					if (t + 5*60 <= now) then
						self.usedBloodlust[guid] = nil
						if (not next(self.usedBloodlust)) then
							self.usedBloodlust = del(self.usedBloodlust)
						end
						freeShaman = true
						whoCanCast = (whoCanCast or "") .. (whoCanCast and ", " or "") .. self:ColourPlayer(unit)
					else
						local cd = t + 5*60
						if (not bloodLustCooldown or cd < bloodLustCooldown) then
							bloodLustCooldown = cd
						end
					end
				else
					freeShaman = true
					whoCanCast = (whoCanCast or "") .. (whoCanCast and ", " or "") .. self:ColourPlayer(unit)
				end
			else
				freeShaman = true
				whoCanCast = (whoCanCast or "") .. (whoCanCast and ", " or "") .. self:ColourPlayer(unit)
			end
		end
	end

	if (shamans == 0) then
		return false, -1

	elseif (satedEnd) then
		if (freeShaman) then
			return false, abs(satedEnd - GetTime()), whoCanCast, 100 - floor(playersWithSated / totalPlayers * 100)
		end
		return false, abs(max(satedEnd, bloodLustCooldown) - GetTime()), whoCanCast, 100 - floor(playersWithSated / totalPlayers * 100)

	else
		if (freeShaman) then
			return true, 0, whoCanCast, 100
		end
		return false, abs(bloodLustCooldown - GetTime()), whoCanCast, 100
	end
end

-- UNIT_SPELLCAST_SUCCEEDED
function module:UNIT_SPELLCAST_SUCCEEDED(ev, unit, spell)
	if (specChangers[spell]) then
		-- They changed talent spec, invalidate their spec and re-fetch
		self:RefreshTalents(unit)

		if (unit == "player") then			-- Don't use UnitIsUnit
			if (self:UserCount() > 1) then
				-- If there's other Utopia users, send my talents immediately
				self:SendMyTalents()
			end
		end

	elseif (spell == GetSpellInfo(Utopia.bloodlustID)) then
		if (UnitInParty(unit) or UnitInRaid(unit)) then
			local guid = UnitGUID(unit)
			if (guid) then
				if (not self.usedBloodlust) then
					self.usedBloodlust = new()
				end
				self.usedBloodlust[guid] = GetTime()
			end
		end
	end
end

-- PLAYER_REGEN_DISABLED
function module:PLAYER_REGEN_DISABLED()
	if (self.frame and db.automode and self.frame.mode ~= "debuffs") then
		if (UnitCanAttack("player", "target")) then
			self.frame:SetMode("debuffs")
		end
	end
end

-- PLAYER_REGEN_ENABLED
function module:PLAYER_REGEN_ENABLED()
	if (self.frame and db.automode and self.frame.mode ~= "buffs") then
		self.frame:SetMode("buffs")
	end
end

do
	-- replaceFunc
	local replaceIndex, thisIndex, replaceValue, replaceColour, replacePercent
	local function replaceFunc(str)
		thisIndex = thisIndex + 1
		if (type(replaceIndex) == "number" and thisIndex == replaceIndex) then
			return replaceValue
		elseif (type(replaceIndex) == "table" and replaceIndex[thisIndex]) then
			return replaceValue
		end
		return str
	end

	local function replaceTalentFunc(str)
		thisIndex = thisIndex + 1
		local num = tonumber(str)
		if (num) then
			if (not replaceIndex or (type(replaceIndex) == "number" and thisIndex == replaceIndex) or (type(replaceIndex) == "table" and replaceIndex[thisIndex])) then
				if (replacePercent) then
					str = replaceColour..tostring(floor(num * replacePercent + 0.5)).."|r"
				elseif (replaceValue) then
					str = replaceColour..tostring(num + replaceValue).."|r"
				end
			end
		end
		return str
	end

	local mmmmBacon = GetSpellInfo(53651)				-- Light's Beacon
	local mmmmCooking = GetSpellInfo(53563)				-- Beacon of Light

	-- FindBacon
	local function FindBacon(unit)
		local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = UnitBuff(unit, mmmmCooking)
		return caster
	end

	-- AddTooltipMapping
	-- *** RECURSIVE ***
	function module:AddTooltipMapping(tooltip, unit, filter, base, name, strrank, tex, count, debuffType, duration, endTime, caster, isStealable)
		local casterName, owner
		if (caster) then
			casterName = self:ColourPlayer(caster)
			if (not UnitIsPlayer(caster) and (UnitPlayerOrPetInRaid(caster) or UnitPlayerOrPetInParty(caster))) then
				-- Is a pet of someone in raid
				local ownerName = self.petlookup and self.petlookup[UnitGUID(caster)]
				if (ownerName) then
					casterName = format("%s <%s>", casterName, self:ColourPlayer(ownerName))
					owner = ownerName
				end
			else
				if (name == mmmmBacon) then
					local source = FindBacon(caster)
					if (source) then
						casterName = format("%s (%s)", casterName, self:ColourPlayer(source))
					end
				end
			end
		end

		-- Add this spell
		local nameDone
		local info = self.lookup[strfind(filter, "HELPFUL") and "buffs" or "debuffs"][name]
		if (not info) then
			info = self.otherBuffs[name]
			if (not info) then
				info = self.classSpells and self.classSpells[name]
			end
		end
		if (info) then
			if (UnitInParty(owner or caster) or UnitInRaid(owner or caster)) then
				local region = _G[tooltip:GetName().."TextLeft2"]
				if (region) then
					local rank
					if (type(strrank) == "string" and strrank ~= "") then
						rank = (tonumber(strrank:match("(%d+)"))) or 1
					end
					local text = region:GetText()

					if (base) then
						if (db.tooltiprank and strrank and strrank ~= "") then
							local rankRegion = _G[tooltip:GetName().."TextRight3"]
							rankRegion:SetText(strrank)
							rankRegion:SetTextColor(0.5, 0.5, 0.5)
							rankRegion:Show()
						end

						if (info.improved) then
							local points = self:HasTalent(owner or caster, info.improved) or 0
							local colour = SmoothColourStr(points / info.maxTalentPoints, true)

							local amount = self:GetAmount(info, nil, nil, owner or caster, name, rank, tex, count, debuffType, duration, endTime, isStealable)
							if ((amount and amount ~= "") or info.regex == "TALENT") then
								--amount = floor(amount)
								local colourAmount = format("%s%s|r", colour, tostring(amount))

								thisIndex = 0
								local regex = info.regex
								if (type(regex) == "boolean") then
									text = text:gsub("(%d+)", colourAmount)
								elseif (type(regex) == "number") then
									replaceIndex = regex
									replaceValue = colourAmount
									text = text:gsub("(%d+)", replaceFunc)
									replaceIndex = nil
									thisIndex = nil
									replaceValue = nil
								elseif (type(regex) == "string") then
									if (regex == "TALENT") then
										replaceIndex = info.regexIndex
										if (info.improvedPercentagePerTalentPoint) then
											replacePercent = 1 + (points * info.improvedPercentagePerTalentPoint / 100)
										elseif (info.improvedAmountPerTalentPoint) then
											replaceValue = floor(points * info.improvedAmountPerTalentPoint)
										end
										replaceColour = colour
										text = text:gsub("(%d+)", replaceTalentFunc)
										replaceColour, replacePercent, replaceValue, replaceIndex = nil, nil, nil, nil
									else
										text = text:gsub(regex, colourAmount)
									end
								elseif (type(regex) == "table") then
									text = text:gsub(regex[1], regex[2]..colourAmount..regex[3])
								end
							end

							local nameString
							local tex = format("Interface\\Addons\\Utopia\\Textures\\%s", (select(2, UnitClass(owner or caster))))
							if (points > 0) then
								-- Add 'Improved 5/5' to end
								nameString = format(L["%s - Improved %s%d|r/%d"], casterName, colour, points, info.maxTalentPoints)
							else
								nameString = format(L["%s - Not Improved"], casterName)
							end
							tooltip:AddLine(nameString, 0.9, 0.8, 0.5)
							tooltip:AddTexture(tex)
							nameDone = true
						end
					else
						local amount = self:GetAmount(info, nil, nil, owner or caster, name, rank, tex, count, debuffType, duration, endTime, isStealable)
						local spellinfo = info.key and self.buffs[info.key]
						if (spellinfo) then
							local descriptor = spellinfo.attributeShort or spellinfo.attribute
							text = format("%s\n|cFFFFFF80%s|r: |cFF80FF80%s|r %s", text, name, amount, descriptor or "")
						end
					end

					if (info.rankLevels) then
						local shouldHaveRank = myLevelRank(info.rankLevels)
						if (info.alternateOffset and info.alternate and name == info.alternate) then
							rank = rank + info.alternateOffset
						end

						if (rank < shouldHaveRank) then
							local shouldHaveAmount = self:GetAmount(info, nil, nil, owner or caster, name, shouldHaveRank, tex, count, debuffType, duration, endTime, isStealable)
							local spellinfo = info.key and self.buffs[info.key]
							if (spellinfo) then
								local descriptor = spellinfo.attributeShort or spellinfo.attribute
								tooltip:AddLine(format(L["This is a rank |cFFFF0000%d|r spell instead of rank |cFFFF0000%d|r which would give you |cFFFFFF80%d %s|r"], rank, shouldHaveRank, shouldHaveAmount, descriptor or ""), 1, 0.4, 0.3, true)
								--tooltip:AddTexture("Interface\\DialogFrame\\DialogAlertIcon")   GameTooltip bugs with a texture on a wrapping line..
							end
						end
					end

					region:SetText(text)
				end
			end
		end

		-- Check mappings for the spell and add those
		local mappings = module.mappingLookup[name]
		if (mappings) then
			for i,mapping in ipairs(mappings) do
				if (not mapping.requiredTalent or (caster and module:HasTalent(owner or caster, mapping.requiredTalent))) then
	--@debug@
					assert(mapping.key ~= name)			-- Would trigger a stack overflow
	--@no-debug@
					self:AddTooltipMapping(tooltip, unit, filter, false, mapping.key, strrank, tex, count, debuffType, duration, endTime, caster, isStealable)
				end
			end
		end

		if (base and not nameDone and caster) then
			tooltip:AddLine(casterName)
			tooltip:AddTexture(format("Interface\\Addons\\Utopia\\Textures\\%s", (select(2, UnitClass(owner or caster)))))
		end
	end

	-- SetUnitAuraExtraInfo
	function module:SetUnitAuraExtraInfo(tooltip, unit, index, filter)
		if (db.tooltip) then
			local name, rank, tex, count, debuffType, duration, endTime, caster, isStealable = UnitAura(unit, index, filter)
			self:AddTooltipMapping(tooltip, unit, filter, true, name, rank, tex, count, debuffType, duration, endTime, caster, isStealable)
			tooltip:Show()
		end
	end
end

-- HookTooltip
function module:HookTooltip()
	self:SecureHook(GameTooltip, "SetUnitAura",
		function(tooltip, unit, index, filter)
			module:SetUnitAuraExtraInfo(tooltip, unit, index, filter)
		end
	)

	self:SecureHook(GameTooltip, "SetUnitBuff",
		function(tooltip, unitid, index)
			module:SetUnitAuraExtraInfo(tooltip, unitid, index, "HELPFUL")
		end
	)

	self:SecureHook(GameTooltip, "SetUnitDebuff",
		function(tooltip, unitid, index)
			module:SetUnitAuraExtraInfo(tooltip, unitid, index, "HARMFUL")
		end
	)
end

-- OpenConfig
function module:OpenConfig(input)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.Utopia)
end

-- OnProfileChanged
function module:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
	self:SetOrientation()
	self:ShowWhen()
end

-- SetupOptions
function module:SetupOptions()
	self.optionsFrames = {}

	-- setup options table
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Utopia", self.options)
	local ACD3 = LibStub("AceConfigDialog-3.0")

	self.optionsFrames.Utopia = ACD3:AddToBlizOptions("Utopia", nil, nil, "General")
	self.optionsFrames.Display = ACD3:AddToBlizOptions("Utopia", L["Display"], "Utopia", "display")
	self.optionsFrames.Icons = ACD3:AddToBlizOptions("Utopia", L["Icons"], "Utopia", "icons")
	self.optionsFrames.ShowWhen = ACD3:AddToBlizOptions("Utopia", L["Show When"], "Utopia", "show")
	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), L["Profiles"])

	-- Add ordering data to the option table generated by AceDBOptions-3.0
	self.options.args.Profiles.order = -2

	self:RegisterChatCommand("utopia", "OpenConfig")
end

-- RegisterModuleOptions
function module:RegisterModuleOptions(name, optionTbl, displayName)
	self.options.args[name] = (type(optionTbl) == "function") and optionTbl() or optionTbl
	self.optionsFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Utopia", displayName, "Utopia", name)
end

-- CheckXPerlTooltipNamesOption
function module:CheckXPerlTooltipNamesOption()
	if (XPerlDB and XPerlDB.buffs) then
		if (db.tooltip) then
			db.wasXPerlNamesOn = XPerlDB.buffs.names
			XPerlDB.buffs.names = nil
		else
			XPerlDB.buffs.names = db.wasXPerlNamesOn
			db.wasXPerlNamesOn = nil
		end
	end
end

-- ToggleDetailsSuper
function Utopia:ToggleDetailsSuper(mode, key)
	if (not Utopia.ToggleDetails) then
		LoadAddOn("Utopia_Details")
	end
	if (Utopia.ToggleDetails) then
		Utopia:ToggleDetails(mode, key)
	end
end

-- OnModuleInitialize
function module:OnInitialize()
	new, del, copy, deepDel = self.new, self.del, self.copy, self.deepDel
	Gradient, UnitFullName, ClassColour, SmoothColour, SmoothColourStr = self.Gradient, self.UnitFullName, self.ClassColour, self.SmoothColour, self.SmoothColourStr
	rotate = Utopia.rotate
	self.potential = {}
	self.talents = {}
	self.label = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
	self.title = Gradient(L["TITLE"].." r"..module.version, 1, 0.1, 0.1, 1, 1, 0.1)

	module.options.name = Gradient(L["TITLE"], 1, 0.1, 0.1, 1, 1, 0.1)
	TalentQuery.RegisterCallback(self, "TalentQuery_Ready")
	if (TalentQuery.events.TalentQuery_Ready_Outsider) then
		TalentQuery.RegisterCallback(self, "TalentQuery_Ready_Outsider")
	end
	playerClass = select(2, UnitClass("player"))
	playerName = UnitName("player")

	SM:Register("border", "Vista", [[Interface\Addons\Utopia\Textures\Border-Vista]])

	self.db = LibStub("AceDB-3.0"):New("UtopiaDB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

	self.db:RegisterDefaults({
		profile = {
			enable = true,
			tooltip = true,
			tooltiprank = true,
			mode = "debuffs",
			show = {
				--always = false,
				--solo = false,
				party = true,
				raid = true,
			},
			popupPos = "top",
			size = 20,
			scale = 1,
			texture = "Blizzard",
			orientation = "H",
			side = "B",
			barwidth = 7,
			hspacing = 0,
			vspacing = 0,
			columns = 15,
			defaultMax = true,
			totemsTemporary = true,
			colour = {
				active = {r = 1, g = 1, b = 1, a = 1},
				inactive = {r = 1, g = 0, b = 0, a = 1},
				unavailable = {r = 1, g = 1, b = 1, a = 0.25},
				partActive = {r = 1, g = 0.8, b = 0.2, a = 1},
				excluded = {r = 0.22, g = 0.62, b = 1, a = 0.75},
				temporary = {r = 0.2, g = 1, b = 0.3, a = 0.75},
			},
			Background = {
				Enable = true,
				Texture = "Solid",
				BorderTexture = "Blizzard Tooltip",
				Color = {r = 0.06, g = 0.06, b = 0.06, a = 0.5},
				BorderColor = {r = 1, g = 1, b = 1, a = 1},
				Tile = false,
				TileSize = 32,
				EdgeSize = 9,
			},
		}
	})
	db = self.db.profile

	self:SetupOptions()
	self:RegisterComm(commPrefix)
	if (self.GetClassSpells) then
		self:GetClassSpells()
	end

	self.OnInitialize = nil
end

-- OnModuleEnable
function module:OnEnable()
	if (self.db) then
		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end

		self:InitData()
		self:OnRaidRosterUpdate()

		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_LEAVING_WORLD")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("PARTY_MEMBER_ENABLE")
		self:RegisterEvent("UNIT_PET")
		self:RegisterEvent("UNIT_LEVEL")
		self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
		self:RegisterEvent("PLAYER_TALENT_UPDATE")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")

		self:CheckXPerlTooltipNamesOption()
		self:HookChat()
		self:HookTooltip()
	end
end

-- OnDisable
function module:OnDisable()
	self:UnhookChat()
	self.lookup.buffs = del(self.lookup.buffs)
	self.lookup.debuffs = del(self.lookup.debuffs)
	self.lookup = del(self.lookup)
	self.buffLookup = del(self.buffLookup)
	self.mappingLookup = del(self.mappingLookup)
	self.classes = deepDel(self.classes)
	self.frame:Hide()
end
